################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

# BuildSystemProvider - an interface that all concrete providers must implement
#
module DevopsBuild
    class BuildSystemProvider
  
      # Starts a new activity for the specified build. Clients must manually
      # complete this activity via BuildSystemProvider#complete_activity.
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * build
      #       The build with which the activity is to be associated
      #    * label (String)
      #       The human-readable string that users will see
      # 
      # Optional
      # --------
      #    * parent_activity
      #        The activity under which this activity should be grouped
      # 
      # Return
      # ======
      #    * activity
      #       An object corresponding to the activity that was started. Clients
      #       should not access any methods or properties on this object but rather
      #       use it as an argument to BuildSystemProvider::complete_activity.
      #
      def start_activity(build, label, parent_activity=nil)
        raise NotImplementedError
      end
  
      # Completes the specified activity
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * activity
      #       The activity to complete
      # 
      def complete_activity(activity)
        raise NotImplementedError
      end
  
      # Publishes an activity for the specified build. Equivalent to calling
      # complete_activity(start_activity(build, label)).
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * build
      #       The build with which the activity is to be associated
      #    * label (String)
      #       The human-readable label that users will see
      # 
      # Optional
      # --------
      #    * parent_activity
      #        The activity under which this activity should be grouped
      def publish_activity(build, label, parent_activity=nil)
        raise NotImplementedError
      end
  
      # Publishes a link to an artifact produced by the specified build.
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * build
      #       The build with which the artifact link is to be associated
      #    * label (String)
      #       The human-readable label that users will see
      #    * url
      #       The url for the published artifact
      #    * local_file_path
      #       The path to the artifact that will be used to compute its size
      # 
      def publish_artifact_link(build, label, url, local_file_path)
        raise NotImplementedError
      end
      
      # Publishes an external URL link.
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * build
      #       The build with which the artifact link is to be associated
      #    * url
      #       The external url
      #    * label (String)
      #       The human-readable label that users will see
      #    * componentName (String)  
      #       The human-readable component name to group links   
      # 
      def publish_link(build, url, label, componentName=nil)
        raise NotImplementedError
      end
      
      # Publishes JUnit test results.
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * build
      #       The build with which the artifact link is to be associated
      #    * junit_result_file
      #       A file with the JUnit test results
      # 
      def publish_junit_results(build, junit_result_file)
        raise NotImplementedError
      end

      # Publishes cloud resources.
      # 
      # Params
      # ======
      #  Required
      #  --------
      #    * build
      #       The build with which the artifact link is to be associated
      #    * instance_id
      #       instanceId on which the application is deployed
      #    * cloud_user_name
      #       user credentials for cloud site
      #    * cloud_password
      #       password credentials for cloud site
      #    * cloud_url
      #       url to the cloud_site
      def publish_cloud_resources(build, instance_id, cloud_user_name, cloud_password, cloud_url)
        raise NotImplementedError
      end
     
      # Publishes cloud logs.
            # 
            # Params
            # ======
            #  Required
            #  --------
            #    * build
            #       The build with which the artifact link is to be associated
            #    * instance_id
            #       instanceId on which the application is deployed
            #    * cloud_user_name
            #       user credentials for cloud site
            #    * cloud_password
            #       password credentials for cloud site
            #    * cloud_url
            #       url to the cloud_site
      def publish_cloud_logs(build, instance_id, cloud_user_name, cloud_password, cloud_url)
        raise NotImplementedError
      end
      
      # Publishes a log file.
      #
      # Params
      # ======
      #  Required
      #  --------
      #    * build
      #       The build with which the artifact link is to be associated
      #    * label 
      #       A label for the logfile. 
      #    * log_file
      #       A filename pointing to the logfile
      #    * status
      #       Status to set in RTC. Valid values are OK, ERROR, and WARNING (default: OK)
      #    * content_type
      #       Content type of the file (default: text/plain)
      def publish_log_file(build, label, log_file, status = 'OK', content_type = 'text/plain')
        raise NotImplementedError
      end

    end # end BuildSystemProvider
end # end module DevopsBuild
