################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "logger" # Ruby Logging

require "httpclient"
require "httpclient/http" # Contains a list of HTTP Codes represented as Ruby Constants
require "open-uri"
require "openssl" 

require 'json' # JSON parsing

require "devops_deployment_lib/util"
require "devops_deployment_lib/error"
require "devops_deployment_lib/http_error"

################################################################################
# Class Definition                                                             #
################################################################################
module DevopsDeploymentLib
  # A class that simplifies the interaction with IWD's Virtual System Pattern REST API.
  #
  class IwdVirtualSystemClient
    include DevopsDeploymentLib::Util

    # HTTP Header Names
    # =================
    HTTP_ACCEPT = "Accept"
    HTTP_CONTENT_TYPE = "Content-Type"
    HTTP_ACCEPT_LANGUAGE = "Accept-Language"
    HTTP_IBM_WORKLOAD_DEPLOYER_API_VERSION = "X-IBM-Workload-Deployer-API-Version"
    #OpenSSL::SSL::VERIFY_PEER = OpenSSL::SSL::VERIFY_NONE unless const_defined?(OpenSSL::SSL::VERIFY_PEER)

    # HTTP Header Values
    # ==================
    APPLICATION_JSON = "application/json"
    IWD_API_VERSION = "3.0"
    LANGUAGE_EN = "en"

    # IWD Resources
    # =============
    IWD_RESOURCES = "resources"
    IWD_VIRTUAL_SYSTEMS = "#{IWD_RESOURCES}/virtualSystems"
    IWD_VIRTUAL_MACHINES = "/virtualMachines"
    IWD_PATTERNS = "#{IWD_RESOURCES}/patterns"
    IWD_SCRIPTS = "#{IWD_RESOURCES}/scripts"
    IWD_VIRTUAL_IMAGES = "#{IWD_RESOURCES}/virtualImages"
    IWD_TEMPLATES = "#{IWD_RESOURCES}/templates"
    IWD_ENV_PROFILES = "#{IWD_RESOURCES}/environmentProfiles"
    IWD_CLOUDS = "#{IWD_RESOURCES}/clouds"
    IWD_IP_GROUPS = "#{IWD_RESOURCES}/ipGroups"
    IWD_VERSION = "#{IWD_RESOURCES}/version"

    # Class Attributes
    # ================
    @iwd_conf
    @http_client
    @log
    @os_deployer_file

    DEPLOYER_CMD = File.join( File.dirname( File.expand_path(__FILE__)), 'deployer_cli', 'deployer.cli', 'bin', 'deployer').to_s
    PURE_CMD = File.join( File.dirname( File.expand_path(__FILE__)), 'deployer_cli', 'pure.cli', 'bin', 'pure').to_s
    ################################################################################
    # Class Methods                                                                #
    ################################################################################
    # Create a new instance of the IwdVirtualSystemClient.
    #
    # == Params
    #
    # === Required
    # [root_server_uri (URI String)]  The URI String that points to an IWD server.
    # [user_name (String)]  The user name for the account that will interact with IWD
    # [password [String]]  The password for the account that will interact with IWD
    #
    # === Optional
    # [:log_level (Logger Constant)]  The log level that the object will output at.
    #                                 Valid values are:
    #                                 * Logger::FATAL
    #                                 * Logger::ERROR
    #                                 * Logger::WARN
    #                                 * Logger::INFO
    #                                 * Logger::DEBUG
    # [:log_file (String|STDOUT|STDERR)]  The file to output logs to
    #
    def initialize(root_server_uri=nil, user_name=nil, password=nil, conf={})
      conf = {
        :log_level => Logger::DEBUG,
        :log_file => 'http_trace.log'
      }.merge(conf)

      @log = Logger.new(conf[:log_file])
      @log.level = conf[:log_level]

      unless root_server_uri.nil? or user_name.nil? or password.nil?

        unless root_server_uri.start_with?("https://")
          # ensure that root_server_uri starts with https://
          root_server_uri = "https://#{root_server_uri}"
        end    

        @debug_str = ""
        @http_client = HTTPClient.new
        @http_client.debug_dev = @debug_str

        @iwd_conf = { # default arguments
          :root_server_uri => root_server_uri,
          :user_name => user_name,
          :password => password,
          :header => {
          HTTP_ACCEPT => APPLICATION_JSON,
          HTTP_ACCEPT_LANGUAGE => LANGUAGE_EN,
          HTTP_IBM_WORKLOAD_DEPLOYER_API_VERSION => IWD_API_VERSION
          }
        }

        set_iwd_user(user_name, password)

      else
        raise ArgumentError, "IwdProvider requires values for root_server_uri, user_name, and password"
      end
    end

    ################################################################################
    # Pattern Interaction Methods                                                  #
    ################################################################################

    # Describes all patterns that are provided by the cloud. If an array of pattern
    # ids are provided, this method will only return data relating to the patterns
    # that those ids represent.
    #
    # == Params
    #
    # === Optional
    # [ids (String Array)]  An array of specific pattern IDs to acquire the data for.
    #
    # == Return
    # [Hash Array]  An array of Hash objects representing the content of the patterns.
    #
    def describe_patterns(args={})
      args = {
        :ids => nil
      }.merge(args)

      if args[:ids].nil?
        return execute_get_request(IWD_PATTERNS)
      else
        instance_arr = []

        args[:ids].each do |id|
          instance = describe_pattern(id)
          instance_arr.push(instance)
        end

        return instance_arr
      end
    end

    # Describes a single pattern that is provided by the cloud.
    #
    # == Params
    #
    # === Required
    # [id (String)]  The ID of a pattern resource
    #
    # == Return
    # [Hash]  An Hash object representing the content of the pattern
    #
    def describe_pattern(id, args={})
      return execute_get_request("#{IWD_PATTERNS}/#{id}")
    end

    # Describes a single pattern part including all the properties of the part.
    #
    # == Params
    #
    # === Required
    # [pattern_id (String)]  The ID of a pattern resource
    # [part_id (String)]  The ID of a part within the pattern resource
    #
    # == Return
    # [Hash]  An Hash object representing the content of the part
    #
    def describe_pattern_part(pattern_id, part_id)
      return execute_get_request("#{IWD_PATTERNS}/#{pattern_id}/pparts/#{part_id}?properties")
    end

    # Describes a template.
    #
    # == Params
    #
    # === Required
    # [template_id (String)]  The ID of a template resource
    #
    # == Return
    # [Hash]  An Hash object representing the content of the template
    #
    def describe_template(template_id)
      return execute_get_request("#{IWD_TEMPLATES}/#{template_id}")
    end

    # Describes scripts that are part of a pattern part.
    #
    # == Params
    #
    # === Required
    # [pattern_id (String)]  The ID of a pattern resource
    # [part_id (String)]  The ID of a part within the pattern resource
    #
    # == Return
    # [Hash]  An Hash object representing the content of the script.
    #
    def describe_part_scripts(pattern_id, part_id)
      return execute_get_request("#{IWD_PATTERNS}/#{pattern_id}/pparts/#{part_id}/pscripts")
    end

    # Describes cloud groups available on an IWD server
    #
    # == Params
    #
    # === Required
    #
    # == Return
    # [Hash]  An Hash object representing the cloud groups
    #
    def describe_cloud_groups
      return execute_get_request("#{IWD_CLOUDS}")
    end

    # Describes environment profiles available on an IWD server
    #
    # == Params
    #
    # === Required
    #
    # == Return
    # [Hash]  An Hash object representing the environment profiles
    #
    def describe_environment_profiles
      return execute_get_request("#{IWD_ENV_PROFILES}")
    end

    # Describes the ICON topologies for a virtual image part if the image was built with ICON
    # otherwise the return is empty.
    #
    # == Params
    # [virtual_image_id (String)]  The ID of the virtual image.
    # [virtual_image_part_id (String)]  The ID of a part within the virtual image resource
    #
    # === Required
    #
    # == Return
    # [Hash]  An Hash object representing the environment profiles
    #
    def describe_virtual_image_topologies(virtual_image_id, virtual_image_part_id)
      return execute_get_request("#{IWD_VIRTUAL_IMAGES}/#{virtual_image_id}/viparts/#{virtual_image_part_id}/vitopologies")
    end

    # Describes all IP groups available on an IWD server
    #
    # == Params
    #
    # === Required
    #
    # == Return
    # [Hash]  An Hash object representing the environment profiles
    #
    def describe_ip_groups
      return execute_get_request("#{IWD_IP_GROUPS}")
    end

    # Describes IWD server version information
    #
    # == Params
    #
    # === Required
    #
    # == Return
    # [Hash]  An Hash object representing the version information
    #
    def describe_version
      return execute_get_request("#{IWD_VERSION}")
    end

    ################################################################################
    # Instance Interaction Methods                                                 #
    ################################################################################

    # Describes all instances accessible by the current user. If an array of instance
    # ids are provided, this method will only return data relating to the instances
    # that those ids represent.
    #
    # == Params
    # === Optional
    # [:ids (String Array)]  An array of instance ids.
    #
    # === Return
    # [Hash Array]  An array containing the Hash objects that represent
    #               the data about all the instances that are accessible
    #               to the current user and filtered by the optional
    #               parameters.
    #
    def describe_instances(args={})
      args = {
        :ids => nil
      }.merge(args)

      if args[:ids].nil?
        return execute_get_request(IWD_VIRTUAL_SYSTEMS, @iwd_conf[:header])
      else
        instance_arr = []

        args[:ids].each do |id|
          instance = describe_instance(id)
          instance_arr.push(instance)
        end

        return instance_arr
      end
    end

    # Describes a single instance that is accessible by the current user.
    #
    # == Params
    # === Required
    # [ids (String)]  The id of an instance.
    #
    # === Return
    # [Hash]  A Hash object representing the data about the requested instance
    #
    def describe_instance(id, args={})
      return execute_get_request("#{IWD_VIRTUAL_SYSTEMS}/#{id}")
    end

    # Get the information about the virtual machines attached to a given instance.
    #
    # == Params
    # === Required
    # [instance_id (String)]  The id of an instance.
    #
    # === Return
    # [Hash]  A Hash object representing the data about the virtual machines attached to
    #         the given instance.
    #
    def describe_instance_virtual_machines(instance_id)
      return execute_get_request("#{IWD_VIRTUAL_SYSTEMS}/#{instance_id}/#{IWD_VIRTUAL_MACHINES}")
    end

    # Get the information about the virtual machines attached to a given
    # instance but without setting the IWD API Client header. This is needed
    # ensure we receive an attribute called 'partKey' that allows us to correlate
    # an VM with a part in a pattern. This requires to fake the user agent.
    #
    # == Params
    # === Required
    # [instance_id (String)]  The id of an instance.
    #
    # === Return
    # [Hash]  A Hash object representing the data about the virtual machines
    # attached to the given instance.
    #
    def describe_instance_virtual_machines_no_header(instance_id)
      args = {}.merge(@iwd_conf)
      # remove IWD header
      args[:header] = {
        HTTP_ACCEPT => APPLICATION_JSON,
        HTTP_ACCEPT_LANGUAGE => LANGUAGE_EN,
        "User-Agent" => "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/535.7 (KHTML, like Gecko) Chrome/16.0.912.75 Safari/535.7"
      }
      response = execute_request(:get, "#{IWD_VIRTUAL_SYSTEMS}/#{instance_id}/#{IWD_VIRTUAL_MACHINES}", HTTP::Status::OK, args)
      resp_data = JSON.parse(response.body)

      return resp_data
    end

    # Get the information about the a given script package
    #
    # == Params
    # === Required
    # [script_id (String)]  The id of a script.
    #
    # === Return
    # [Hash]  A Hash object representing the data about the script.
    #
    def describe_script_package(script_id)
      return execute_get_request("#{IWD_SCRIPTS}/#{script_id}")
    end

    # Get the information about the a given virtual image
    #
    # == Params
    # === Required
    # [script_id (String)]  The id of a virtual image.
    #
    # === Return
    # [Hash]  A Hash object representing the data about the virtual image.
    #
    def describe_virtual_image(image_id)
      return execute_get_request("#{IWD_VIRTUAL_IMAGES}/#{image_id}")
    end

    # Create new instances from Virtual System Patterns (VSP).
    #
    # == Params
    # == Required
    # [instance_attributes (Hash)]  A Hash object containing all of data required to create a new
    #                               instance of the VSP.
    #
    # == Return
    # [Hash]  A Hash object containing the data that represents the newly created VSP instance.
    #
    def create_instances(instance_attributes, args={})
      execute_post_request(IWD_VIRTUAL_SYSTEMS, instance_attributes)
    end

    # Terminates all instances that are provided.
    #
    # == Params
    #
    # === Required
    # [ids (String Array)]  An array of instance ids
    #
    # == Return
    # [Hash Array]  An array of Hash objects containing the data of all the instances that were
    #               terminated
    #
    def terminate_instances(ids=[], args={})

      if ids.nil? or ids.length < 1
        raise ArgumentError "Must provide one or more ids"
      else
        instance_arr = []

        ids.each do |id|
          instance = execute_delete_request("#{IWD_VIRTUAL_SYSTEMS}/#{id}")
          instance_arr.push(instance)
        end

        return instance_arr
      end
    end

    # Create a script package in the catalog
    def create_script_package(script_package_name, archive_file, replace)
      configure_cli() # if needed
      args = []
      args << " --name=\"#{script_package_name}\" "
      args << " --file=\"#{archive_file}\" "
      if replace
        args << " --replace "
      end
      create_script_package_py = File.join( File.dirname( File.expand_path(__FILE__)), 'jython', 'createScriptPackage.py')
      result = exec_jython(create_script_package_py, args)
      # return result??
    end

    # Create a virtual system pattern in the catalog
    def create_pattern(pattern_name, virtual_image_name, part_name)
      configure_cli() # if needed
      args = []
      args << " --name=\"#{pattern_name}\" "
      args << " --partName=\"#{part_name}\" "
      args << " --imageName=\"#{virtual_image_name}\" "
      create_pattern_py = File.join( File.dirname( File.expand_path(__FILE__)), 'jython', 'createPattern.py')
      result = exec_jython(create_pattern_py, args)
      @log.debug "result of create_pattern: #{result}"
      # return result??
    end

    # Create a virtual system pattern in the catalog
    def add_script_package_to_pattern(pattern_name, part_name, script_package_name)
      configure_cli() # if needed
      args = []
      args << " --name=\"#{pattern_name}\" "
      args << " --partName=\"#{part_name}\" "
      args << " --scriptPackage=\"#{script_package_name}\" "
      add_script_package_to_pattern_py = File.join( File.dirname( File.expand_path(__FILE__)), 'jython', 'addScriptPackageToPattern.py')
      result = exec_jython(add_script_package_to_pattern_py, args)
      @log.debug "result of add_script_package_to_pattern: #{result}"
      # return result??
    end
    
    # Create a virtual system pattern in the catalog
    def grant_user_permission_to_instance(instance_id)
      configure_cli() # if needed
      virtualsysurl = "#{@iwd_conf[:root_server_uri]}/dashboard/runtime/virtual/##{instance_id}"
      @log.debug 
      @log.debug "Cloud user_id = #{@iwd_conf[:user_name]}"
      @log.debug "Cloud virtualsysurl == #{virtualsysurl}"
      args = []
      args << " #{@iwd_conf[:user_name]}"
      args << " #{virtualsysurl}"
      add_script_package_to_pattern_py = File.join( File.dirname( File.expand_path(__FILE__)), 'jython', 'grantUserPermissionToInstance.py')
      result = exec_jython(add_script_package_to_pattern_py, args)
      @log.debug "result of grant_user_permission_to_instance: #{result}"
      # return result??
    end
    ################################################################################
    # Helper Methods                                                               #
    ################################################################################

    # Set the user account to interact with IWD.
    #
    # == Params
    #
    # === Required
    # [user_name (String)]  The user name for the account that will interact with IWD
    # [password [String]]  The password for the account that will interact with IWD
    #
    def set_iwd_user(user_name, password)
      @http_client.set_auth(nil, user_name, password)
      @http_client.www_auth.basic_auth.challenge(@iwd_conf[:root_server_uri])
    end

    # Set the SSL verify mode of the HTTP client that will interact with the IWD system.
    #
    # == Params
    #
    # === Required
    # [verify_mode (OpenSSL Constant)]  The verify mode that should be used to interact
    #                                   with servers over https. Valid values are:
    #                                   * OpenSSL::SSL::VERIFY_NONE
    #                                   * OpenSSL::SSL::VERIFY_PEER
    #                                   * OpenSSL::SSL::VERIFY_CLIENT_ONCE
    #                                   * OpenSSL::SSL::VERIFY_FAIL_IF_NO_PEER_CERT
    #
    def ssl_verify_mode=(verify_mode)
      @http_client.ssl_config.verify_mode = verify_mode
      @iwd_conf[:ssl_verify_mode] = verify_mode
    end

    private
    ################################################################################
    # Private Methods                                                              #
    ################################################################################

    # HTTP Request Execution

    def execute_get_request(resource, args={})
      args = {
        :header => {}.merge(@iwd_conf[:header])
      }.merge(args)

      response = execute_request(:get, resource, HTTP::Status::OK, args)
      resp_data = JSON.parse(response.body)

      return resp_data
    end

    def execute_post_request(resource, body, args={})
      args = {
        :body => body,
        :header => {
        HTTP_CONTENT_TYPE => APPLICATION_JSON
        }.merge(@iwd_conf[:header])
      }.merge(args)

      response = execute_request(:post, resource, HTTP::Status::CREATED, args)
      resp_data = JSON.parse(response.body)

      return resp_data
    end

    def execute_put_request(resource, body, args={})
      args = {
        :body => body,
        :header => {
        HTTP_CONTENT_TYPE => APPLICATION_JSON
        }.merge(@iwd_conf[:header])
      }.merge(args)

      response = execute_request(:put, resource, HTTP::Status::OK, args)
      resp_data = JSON.parse(response.body)

      return resp_data
    end

    def execute_delete_request(resource, args={})
      args = {
        :header => {}.merge(@iwd_conf[:header])
      }.merge(args)

      response = execute_request(:delete, resource, HTTP::Status::NO_CONTENT, args)
      resp_data = JSON.parse(response.body)

      return resp_data
    end

    def execute_request(http_method, resource, expected_status, args={})
      response = nil

      case http_method
      when :get
        response = @http_client.get("#{@iwd_conf[:root_server_uri]}/#{resource}", :query => args[:query], :header => args[:header])
      when :post
        response = @http_client.post("#{@iwd_conf[:root_server_uri]}/#{resource}", :body => args[:body], :header => args[:header])
      when :delete
        response = @http_client.delete("#{@iwd_conf[:root_server_uri]}/#{resource}", :header => args[:header])
      when :put
        response = @http_client.put("#{@iwd_conf[:root_server_uri]}/#{resource}", :body => args[:body], :header => args[:header])
      end

      @log.debug @debug_str

      if response.status == expected_status
        return response
      else
        raise HttpError.new("Error: Unrecognized Code, #{response.code} - #{response.reason}", response.code)
      end

    end

    def configure_cli()
      unless File.exists?(DEPLOYER_CMD) or File.exists?(PURE_CMD)
        @log.info "File does not exist: #{DEPLOYER_CMD} or #{PURE_CMD}. Configuring Jython CLI."
        cli_url = "#{@iwd_conf[:root_server_uri]}/downloads/cli"
        cli_filename = File.join( File.dirname( File.expand_path( __FILE__ )), 'iwd_cli.zip')
        deployer_cli = File.join( File.dirname( File.expand_path( __FILE__ )), 'deployer_cli')
        @log.debug "Fetching CLI into #{cli_filename} from #{cli_url}"
        unless File.exists?(cli_filename)
          get_file(cli_filename, cli_url, @iwd_conf[:ssl_verify_mode] ) # Requires Ruby 1.9, :ssl_verify_mode => @iwd_conf[:ssl_verify_mode]
          raise Exceptions::Failure, "Could not download the cloud command Line interface from \"#{cli_url}\" to \"#{cli_filename}\". Please verify that the host and target file are each accessible." unless File.exists? cli_filename
        end
        @log.debug "Unzipping CLI into #{deployer_cli}"
        unzip_file(cli_filename, deployer_cli)
        if RUBY_PLATFORM.downcase != "i386-mingw32"
          `chmod -R 777 #{deployer_cli}`
        end
      end  

      if File.exists?(DEPLOYER_CMD)
        @log.info "Using IBM SmartCloud Provisioning API: #{@os_deployer_file}"
        @os_deployer_file = DEPLOYER_CMD
      elsif File.exists?(PURE_CMD)
        @log.info "Using PureApplication API: #{@os_deployer_file}"
        @os_deployer_file = PURE_CMD
      else
        raise Exceptions::Failure, "Could not find cloud command  \"#{DEPLOYER_CMD}\" or \"#{PURE_CMD}\". Please verify that you have the correct permissions to extract the cloud command line archive."
      end

      if (@os_deployer_file.match(" "))
        @os_deployer_file = "\"#{@os_deployer_file}\""
      end
      if @os_deployer_file.nil? or @os_deployer_file.empty? 
        raise Exceptions::Failure, "Could not find cloud command  \"#{DEPLOYER_CMD}\" or \"#{PURE_CMD}\". Please verify that you have the correct permissions to extract the cloud command line archive."
      end
    end

    def exec_jython(script, args=[])

      if @iwd_conf[:root_server_uri] && @iwd_conf[:user_name] && @iwd_conf[:password] && script

        hostname = @iwd_conf[:root_server_uri]
        if(hostname.start_with?("https://"))
          hostname = hostname["https://".length, hostname.length]
        end

        if (hostname.include? ':')
          port = hostname.split(':')[1]
          hostname = hostname.split(':')[0]
          command = "#{@os_deployer_file} -h #{hostname} -u #{@iwd_conf[:user_name]} -p #{@iwd_conf[:password]} -P #{port}"
        else
          command = "#{@os_deployer_file} -h #{hostname} -u #{@iwd_conf[:user_name]} -p #{@iwd_conf[:password]}"
        end

        if(File.file?(script))
          command = command + " -f \"#{script}\" "
        end
        args.each do |arg|
          command = command + "  #{arg}"
        end
        command = command + " >> deployer_cli.log"

        @log.debug "exec_jython command string: \"#{command}\""
        # why system() instead of backtick?

        @log.debug "exec_jython command: \"#{command}\""
        output = `#{command}`
        # `#{command}` doesn't return properly
        @log.debug "exec_jython command output: #{$?}"
        @log.debug "result of exec_jython #{script}: \"#{output}\" - #{$?}"
        unless $? == 0
          raise Exceptions::Failure, "Command execution failed with code #{$?} using \"#{command}\". See \"deployer_cli.log\" for more information."
        end 
        
      else
        @log.warn "exec_jython could not execute script! #{@iwd_conf[:root_server_uri]} && #{@iwd_conf[:user_name]} && #{@iwd_conf[:password]} && #{script}"
      end
    end

  end # end IwdVirtualSystemClient class
end # end DevopsDeploymentLib module
