require 'weaver/cli/weaver_cli_plugin'
require 'weaver/chef/chef_node_inspector'
require 'weaver/chef/chef_automation_handler'
require 'fileutils'
require 'erb'

module Weaver
  module Cli
    module Plugins
      # \Chef plugin for the Weaver CLI
      #
      # See http://www.opscode.com/chef/ for information on \Chef itself
      #
      # Syntax: <tt>weaver chef [task] [options]</tt>
      #
      # Author:: Ed Snible
      #
      class ChefCliPlugin < Weaver::Cli::WeaverCliPlugin

        # Special non-Thor extension to provide help for the arguments of the commands
        ARG_DEFS = {
          'ROLE_FILE' => 'A Chef Role file (*_role.rb)',
          'ENVIRONMENT_FILE' => 'A .weaver file defining an environment',
        }
      
        namespace :chef
        namespace_description "Chef related tasks"

        desc "import ROLE_FILE", "Create a Weaver description of a Chef Role file (*_role.rb) Chef recipe (cookbook::recipe)"
        method_option :cookbooks, :type => :string, :desc => "Colon-separated folders named 'chef' containing Chef setup."
        method_option :loc, :type => :string, :aliases => "-l", :desc => "Location for the resulting .weaver file."
        # Create a \Weaver description of a \Chef role file (*_role.rb)
        # * +file+ - \Chef role file (*_role.rb)
        # * +cookbooks+ - Colon-separated folders named 'chef' containing \Chef setup.
        # * +loc+ - Location for the resulting .weaver file.
        def import(file = nil)
          collect_properties
          setup_logging

          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"import" was called incorrectly. Call as "import ROLE_FILE" or "import cookbook::recipe".}) unless file

          # 'chef_path' should be something like "~/src/eclipse/devops/com.ibm.devops.sample.infrastructure/chef:~/src/eclipse/devops/com.ibm.devops.common.infrastructure/chef"
          chef_path = options[:cookbooks] || ENV['CHEF_SETUP_PATHS']
          say_status("Error", "Parameter 'cookbooks' or environment variable 'CHEF_SETUP_PATHS' not set.", :red) && exit(1) unless chef_path

          unless chef_path.nil?
            chef_path.split(File::PATH_SEPARATOR).each do |dirname|
              unless File.directory?(dirname)
                say_status("Error", "Directory #{dirname} does not exist", :red)
              else
                say_status("Error", "#{dirname} is not a Chef directory (does not contain 'cookbooks' or 'roles' subdirectories)", :red) unless
                File.directory?(File.expand_path('cookbooks', dirname)) || File.directory?(File.expand_path('roles', dirname))
              end
            end
          end

          inspector = Weaver::Chef::ChefNodeInspector.new(chef_path.split(File::PATH_SEPARATOR)) do |level, msg|
            report_message("Importing", level, msg)
          end

          dir = options[:loc] || '.' # where to store it
          
          # first, see if we need to auto-generate the role file
          file = auto_generate_role_if_needed(file, chef_path)

		  # next, actually generate the .weaver file
          begin
            model = inspector.model_role(file, dir)
            extension = ".rb"
          rescue WeaverError => we
            say_status("Error", "Chef introspection failed. Original message: '#{we.to_s}'", :red)
            exit(1)
          end

          name = File.basename(file, extension)
          full_filename = filename = name + Weaver::Core::WEAVER_SUFFIX

          if options[:loc]
            full_filename = File.join(dir, full_filename)
          end

          say_status("Saving", "Role definition '#{filename}' in '#{dir}'", :green)

          # creating a backup copy if file already exists
          if File.exists?(full_filename)
            say_status("Creating:", "File '#{filename}' already exists. Creating backup '#{filename}.saved'", :green)
            FileUtils.copy(full_filename, "#{full_filename}.saved")
          end

          FileUtils.mkdir_p(File.dirname(full_filename)) unless File.directory?(File.dirname(full_filename))
          file = File.open(full_filename, 'w')
          File.open(file.path, 'w') do |f|
            f.write( model.to_weaver )
          end

        end

        desc "nodes ENVIRONMENT_FILE", "Generates the Chef node files (for chef-solo) for a given Weaver environment"
        method_option :loc, :type => :string, :aliases => "-l", :desc => "Location for the resulting Chef node file."

        # Create a \Weaver description of a \Chef role file (*_role.rb)
        # * +environment_file+ - \Weaver environment file
        def nodes(environment_file = nil)
          collect_properties
          setup_logging
          
          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"nodes" was called incorrectly. Call as "nodes ENVIRONMENT_FILE".}) unless environment_file
          
          if options[:loc]
            dir = File.expand_path(options[:loc])                    
          else 
            dir = "."
          end
          
          if File.exists?(environment_file) && File.directory?(environment_file)
            say_status("Error", "File '#{environment_file}' is a directory.", :red)
            exit(1) 
          end
          if !File.exists?(environment_file) 
            say_status("Error", "File '#{environment_file}' does not exist.", :red)
            exit(1)
          end          
          if !File.directory?(dir)
            say_status("Error", "Directory '#{dir}' is not a valid directory.", :red)
            exit(1)
          end
          
          handler = Weaver::Chef::ChefAutomationHandler.new() { |level, msg| report_message("Generating", level, msg) }            
          begin
            # This should use Environment here as it implements the logic already but this CLI function would require licensing then.
            root = Weaver::Dsl::WeaverParser.new().parse(environment_file)
            topology = root.topologies.values.to_a[0]
            if root.topologies.size > 1
              @@logger.warn "#{environment_topology_file} contains #{@root.topologies.size} top-level topologies, using #{@topology.id.inspect}"
            end
            if topology.nil?
              @@logger.error "No topology in #{environment_topology_file.inspect}"
            end
            topology = Weaver::Transforms::Fold.fold(topology)

            handler.generate_node_files(topology, dir)
            say_status("Done", "chef-solo node files generated successfully.", :green)

          rescue WeaverError => we
            say_status("Error", "chef-solo node file generation failed. Original message: '#{we.to_s}'", :red)
            exit(1)
          end
        end

        no_tasks do
          # @private
          # :stopdoc:
          # Report a status message running a command
          def report_message(status, level, msg)
            if level == Logger::ERROR || level == Logger::FATAL
              say_status(status, msg, :red)
            elsif level == Logger::WARN
              say_status(status, msg, :yellow)
            else
              say_status(status, msg, :green)
            end
          end
          
          # @private
          # :stopdoc:
          # The file_or_recipe param can either be the name of a role file, or 
          # cookbook::recipe.  If file_or_recipe is in the format cookbook::recipe, 
          # find the file roles/cookbook_recipe.rb, or auto-generate it if needed.
          # Returns the name of the role file.
          def auto_generate_role_if_needed(file_or_recipe, chef_path)
            
            if(file_or_recipe.nil? || file_or_recipe.end_with?(".rb"))
            	# do not process
            	return file_or_recipe
            end
            
            begin
            	cookbook_name = file_or_recipe.split("::")[0]
            	recipe_name = file_or_recipe.split("::")[1]
            rescue Error => e
            	say_status("Error", "Chef role file generation failed.  Call as \"import ROLE_FILE\" or \"import cookbook::recipe\"", :red)
            	exit(1)
            end
            
            role_name = "#{cookbook_name}_#{recipe_name}"
            role_file_shortname = "#{role_name}.rb"
            
            # Search for an existing role_file for the recipe
            unless chef_path.nil?
            	chef_path.split(File::PATH_SEPARATOR).each do |dirname|
            		# This logic will try to find the role_file in each directory
              		if File.directory?(dirname)
              			if File.directory?(File.expand_path('cookbooks', dirname)) && File.directory?(File.expand_path('roles', dirname))
              				# look for roles/role_name.rb
              				role_file = File.join( File.expand_path(dirname), "roles", role_file_shortname)
              				if File.exist?(role_file)
              					return role_file_shortname # we found it, return the existing file
              				else
              					# make sure the recipe actually exists
              					recipe_file = File.join( File.expand_path(dirname), "cookbooks", cookbook_name, "recipes", "#{recipe_name}.rb")
              					unless File.exist?(recipe_file)              					
              						say_status("Error", "#{recipe_file} does not exist.  Cannot automatically generate role file #{role_file}.  Check the cookbook name and recipe name, or create a role file, and try again.", :red)
              						exit(1)
              					end
              					# generate the file
            					role_description = "Runs the #{recipe_name} recipe from cookbook #{cookbook_name}"
            						
            					template_file = File.join( File.dirname(__FILE__), "..", "..", "..", "..", "data", "templates", "chef_role.erb")
              					if File.exist?(template_file)
              						# generate the role file
          							say_status("Saving", "Automatically generating role definition '#{role_file}'", :green)
              						renderer = ERB.new File.new(template_file).read, nil, "%"
									output = renderer.result(binding)
              						
    								File.open(role_file, "w+") do |f|
      									f.write(output)
    								end
              					else
              						say_status("Error", "#{template_file} does not exist.  Cannot automatically generate role file #{role_file}.  Create the role file and try again.", :red)
              						exit(1)
              					end
            					
            					return role_file_shortname
              				end
              			else # if it gets here, cookbooks or roles subdirectories are missing
                			say_status("Error", "#{dirname} is not a Chef directory (does not contain 'cookbooks' and 'roles' subdirectories)", :red)
              			end # end dirname contains cookbooks or roles subdirectories
              		else # if it gets here, dirname is not a directory
                		say_status("Error", "Directory #{dirname} does not exist", :red)
                	end
              	end # end processing each dirname in chef_path
            end # chef_path is not nil
            
          	return file_or_recipe
          end
          # :startdoc:
          
        end

      end
    end

  end
end