require 'string'
require 'timeout'
require 'hash'
require 'weaver/cli/weaver_cli_plugin'
require 'weaver/dsl/weaver_parser'
require 'weaver/dsl/root_statement'
require 'weaver/generators/environment_generator'
require 'weaver/transforms/fold'
require 'weaver/transforms/unroll'
require 'weaver/environment'
require 'weaver/system'
require 'weaver/service/service_client'
require 'weaver/service/weaver_service_error'

module Weaver
  module Cli
    # \Plugins provides plug-ins to be used by Cli.
    module Plugins
      # Environment plugin for the Weaver CLI
      #
      # Syntax: weaver env [task] [options]
      #
      # Author:: Florian Rosenberg
      # Author:: Ed Snible
      #
      class EnvCliPlugin < Weaver::Cli::WeaverCliPlugin

        # Special non-Thor extension to provide help for the arguments of the commands
        ARG_DEFS = {
          'WEAVER_FILE' => 'Any .weaver file',
          'EXPR' => 'A valid Ruby expression to evaluate in the context of the Weaver file',
          'ENVIRONMENT_ID' => 'A name for the generated Environment',
        }

        namespace :env
        namespace_description "Environment related tasks"

        desc "eval WEAVER_FILE EXPR", "Evaluate a Weaver expression in the context of a file"
        method_option :config, :required => false, :type => :hash, :aliases => "-c", :desc => "Configuration parameters.", :default => {}
        method_option :deploy_params, :required => false, :type => :string, :aliases => "-d", :desc => "Deployment parameters.  DEPRECATED"

        # Evaluate an expression against an Environment.  Used for debugging.
        def eval(weaver_file = nil, expr = nil)
          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"eval" was called incorrectly. Call as "eval WEAVER_FILE EXPR".}) unless weaver_file && expr

          loader = init(options[:config])
          if weaver_file.nil?
            weaver_file = loader.default_environment
          end
          validate_file_exist?(weaver_file)
          deploy_params = loader.deployment_parameters(parse_deploy_parameters(options[:deploy_params]))

          wp = Weaver::Dsl::WeaverParser.new(loader.config_params)
          root = wp.parse(weaver_file)
          begin
            e =  Weaver::Environment.new(root, loader) { |lvl, status, msg| weaver_report(lvl, status, msg) }
            root = e.simulate_deploy({}, deploy_params)
          rescue WeaverError => we
            @@logger.warn "Unable to create environment: #{we.message}; backtrace: #{we.backtrace}"
            say_status("Error", "Unable to create environment: #{we.message}", :red)
            return
          end

          begin
            say_status("#{expr.inspect}", root.instance_eval(expr).inspect, :green)
          rescue Exception => ex
            @@logger.warn "Unable to evaluate: #{ex.message}; backtrace: #{ex.backtrace}"
            say_status("Error", "Unable to evaluate: #{ex.message}", :red)
            exit(1)
          end
        end

        desc "validate WEAVER_FILE", "Validate a .weaver file."
        method_option :config, :required => false, :type => :hash, :aliases => "-c", :desc => "Configuration parameters.", :default => {}
        method_option :deploy_params, :required => false, :type => :string, :aliases => "-d", :desc => "Deployment parameters. DEPRECATED"
        method_option :phases, :type => :string, :aliases => "-p", :required => false, :desc => "Comma-separated list of validation phases. Valid phases are 'predeployment' and 'postdeployment'."
        method_option :exclude, :type => :string, :required => false, :desc => "Comma-separated list of tags associated with rules to be ignored during validation."

        # Parse a Weaver file and print messages if the file is not valid.  If the Weaver file seems to be an environment it will be folded first.
        def validate(weaver_file = nil)
          loader = init(options[:config])

          # Validate the default environment file by load it from config if not given as CLI argument
          if weaver_file.nil?
            weaver_file = loader.default_environment
          end
          validate_file_exist?(weaver_file)

          # FIXME this needs to be moved into the validate API call??
          rules =  File.join(Dir.pwd, 'rules')

          if !options[:no_color]
            success_color = :green
            info_color = :clear
            warn_color = :yellow
            error_color = :red
          else
            success_color = info_color = warn_color = error_color = nil
          end

          say_status("Validating", "File '#{weaver_file}'", :green)

          deploy_params = loader.deployment_parameters(parse_deploy_parameters(options[:deploy_params]))
          wp = Weaver::Dsl::WeaverParser.new(loader.config_params)
          begin
            root = wp.parse(weaver_file)
          rescue WeaverError => we
            @@logger.warn "Fatal error #{we.message} parsing '#{weaver_file}'; backtrace: #{we.backtrace}"
            say_status("Error", "Fatal error #{we.message} parsing '#{weaver_file}'", :red)
            return
          rescue Exception => ex
            @@logger.debug "#{ex}, backtrace=#{ex.backtrace}"
            say_status("Error", "Unable to parse Weaver file: #{ex.message}", :red)
            return
          end
          
          phases = options[:phases]
          # parse phases
          parsed_phases = []
          unless phases.nil?
            phases.split(',', -2).each do |item|
              aux = item.upcase
              if aux =~ /\APREDEPLOYMENT\Z$/ || aux =~ /\APOSTDEPLOYMENT\Z$/
                parsed_phases << aux.swapcase.to_sym
              else
                say_status("Error:", "Invalid validation phase. Valid phases are 'predeployment' and 'postdeployment'. The list must be comma-separated.", :red)
                exit
              end
            end
          end

          msg = "Validation will consider the following phase(s): "
          parsed_phases.each do |p|
            msg += "#{p} "
          end

          exclude = options[:exclude]
          # parse tags whose associated rules are to be excluded from validation
          parsed_exclude = []
          unless exclude.nil?
            parsed_exclude = exclude.split(',').map {|item| item.to_sym if item != "" && !item.nil?}
            parsed_exclude.delete_if {|item| item.nil?}
          end
          @@logger.debug "Considering the following exclude tags: #{parsed_exclude.inspect}"

          unless rules.nil?
            rules_files = files(rules + '/*.weaver')
            #say "No rules files found" if rules_files.size == 0
            rules_files.each do |file|
              say "Processing #{file.path}" if options[:verbose]
              # Create a new root Statement for parsing error printing, but overlayed onto the
              # root of the file we are to validate
              root_statement = Weaver::Dsl::RootStatement.new(file.path, wp)
              root_statement.model = root

              wp.parse(file, root_statement)
            end
          end

          out_messages = []
            
#          # If the Weaver document to validate is an Environment we shall ask the Environment class to prepare it
#          # exactly as would happen for deployment.  This allows the environment to invoke Weaver::Deployers::Deployer s
#          # (for example, to validate fixed IP address requests), and to generate attributes such as elastic IP addresses
#          # that will be propagated and validated.
#          unless unfolded
            # Skip simulating deployment unless we have at least one provider and are not an application topology
            unless root.flatten_children.select { |child| child.is_a?(ProviderContainerMixin) && child.providers.size > 0 }.empty?
              begin
                e = Weaver::Environment.new(root, loader) { |lvl, status, msg| weaver_report(lvl, status, msg) }
                root = e.simulate_deploy(loader.config_params, deploy_params).root                
              rescue WeaverError => we
                @@logger.debug "#{we}, backtrace=#{we.backtrace}"
                # We couldn't create the environment; treat this as a validation problem
                root.add_validation_failure(we.message)
              rescue StandardError => se
                @@logger.debug "#{se}, backtrace=#{se.backtrace}"
                # For example, Errno::EACCES: Permission denied or RuntimeError
                say_status("Error", se.message, :red)

                # We couldn't create the environment; treat this as a validation problem
                root.add_validation_failure(se.message)
              end
            else
              # We have an application, but fold what we do have anyway, to check for realization problems
              begin
                root = Weaver::Transforms::Fold.fold(root)
              rescue WeaverError => we
                root.add_validation_failure(we.message)
              end
              
              # Add in problems with the config.yml since Environment can't do it on our behalf
              loader.validation_messages.each { |msg| root.add_validation_failure(msg) }
            end

          begin
            root.validate_collect(out_messages, parsed_phases, parsed_exclude)
          rescue Exception => ex
            @@logger.warn "Fatal error #{ex.message} validating '#{weaver_file}'; backtrace: #{ex.backtrace}"
            say_status("Error", "Unable to validate because of internal error: #{ex.message}", :red)
            exit(1)
          end

          if out_messages.size == 0
            say_status("Done", 'Validation successful', :green)
          else
            out_messages.map { |msg| msg.to_s }.each_with_index do |message, i|
              message = message.to_s
              if / ERROR / =~ message || / FATAL / =~ message
                color = error_color
              elsif / WARNING / =~ message
                color = warn_color
              else
                color = info_color
              end
              if options[:no_color]
                say("#{message}", color)
              else
                say("#{i+1}) #{message}", color)
              end
            end
          end
        end

        desc "deploy ENVIRONMENT_FILE", "Deploy an environment to a cloud."
        method_option :config, :required => false, :type => :hash, :aliases => "-c", :desc => "Configuration parameters.", :default => {}
        method_option :deploy_params, :required => false, :type => :string, :aliases => "-d", :desc => "Deployment parameters. DEPRECATED"
        method_option :timeout, :type => :numeric, :aliases => "-o", :desc => "Timeout for the deployment request (in seconds)"

        # Deploy an IAAS Pattern modeled as a Weaver file
        def deploy(environment_file = nil)
          loader = init(options[:config])

          # load environment from config if not given as CLI argument
          if environment_file.nil?
            environment_file = loader.default_environment
          end
          validate_file_exist?(environment_file)

          deploy_params = loader.deployment_parameters(parse_deploy_parameters(options[:deploy_params]))
          properties = loader.config_params.merge({ :log_level => @@logger.level, :timeout => options[:timeout]})

          # TODO INTEGRATION: comment this out in integration stream
          if (use_service)
            deploy_with_service(environment_file, properties, deploy_params)
            return
          end

          say_status("Deploying", "Environment '#{environment_file}'", :green)

          begin
            @@logger.debug("Deployment parameters: #{deploy_params}")
            e =  Weaver::Environment.new(environment_file, loader) { |lvl, status, msg| weaver_report(lvl, status, msg) }

            # templates are not available in the INTGRATION stream; should be false
            if use_templates           
              e.deploy_or_update(:deploy, {}, nil, properties, deploy_params)
            else
              e.deploy(properties, deploy_params)
            end

            # Print results
            if e.root.topologies.values.first.amd[:deployment]
              say_status("Deployment id", "#{e.root.topologies.values.first.amd[:deployment].value} [Status: #{e.root.topologies.values.first.amd[:config_status].value}]", :green)
            end

            say_status("Done", "Summary of deployed nodes:", :green)
            table = []
            table << ['Node ID', 'Hostname', 'IP Address', 'Status']
            table << ['-------', '--------', '----------', '-------']
            # TODO this needs to be simplified
            e.topology.flatten_children.select { |n| n.respond_to?(:realized_by) && n.realized_by.empty? && !n.tags.member?(:storage) && !Weaver::Transforms::Unroll.template?(n)}.each do |v|
              config_status = v.amd[:config_status].nil? ? "unknown" : "#{v.amd[:config_status].value}"
              #say_status("Node '#{v.id.to_s}'", "#{v.hostname} - #{v.ip_address}   #{config_status}", :green)
              table << ["#{v.id.to_s}", "#{v.hostname}", "#{v.ip_address}", "#{config_status}"]
            end
            say() # empty line
            print_table(table, {:indent => 14})

          rescue Timeout::Error => te
            @@logger.warn "Fatal error #{te.inspect} parsing '#{environment_file}'; backtrace: #{te.backtrace.join("\n")}"
            say_status("Error", "Timeout exceeded. Please use the '--timeout' option to increase the deployment timeout.",  :red)
            exit(1)
          rescue WeaverError => ex
            @@logger.warn "Fatal error #{ex.inspect} parsing '#{environment_file}'; backtrace: #{ex.backtrace.join("\n")}"
            if ex.original.kind_of?(Timeout::Error) # for some reason Timeout:Error get wrapped into WeaverError in 1.9
              say_status("Error", "Timeout exceeded. Please use the '--timeout' option to increase the deployment timeout.",  :red)
            else
              say_status("Error", ex.message, :red)
            end
            exit(1)
          rescue RuntimeError => re
            @@logger.warn "Fatal error #{re.inspect} parsing '#{environment_file}'; backtrace: #{re.backtrace.join("\n")}"
            say_status("Error", re.to_s, :red)
            exit(1)
          rescue Exception => e
            @@logger.warn "Fatal error #{e.inspect} parsing '#{environment_file}'; backtrace: #{e.backtrace.join("\n")}"
            say_status("Error", "Failed to deploy #{e.inspect}",  :red)
            exit(1)
          end
        end

        # location for writing deployed instances when update used (temporary)
        DEPLOYMENT_DIR ="tmp/my_instances"
        
        # CLI command to list delployed environment instances.  When called from  weaver project folder, 
        # returns those deployments created by/fromdd
        desc "list", "List deployed instances of the environment"
        def list
          begin 
#            create_coordinator()
#            environments = Weaver::Coordination::CoordinatedTransactionManager.instance.list
            return unless File.directory? DEPLOYMENT_DIR
            environments = Dir.foreach("tmp/my_instances").inject({}) do |hash, e|
              env = nil
              begin
                env = instance_model(e)
              rescue
                env = nil
              end
                hash[e] = env unless env.nil?
                hash
            end
            environments.each do |id, env|
              say_status("Deployment: ", "#{id}", :green)
            end
          rescue Exception => e
            say_status("Error", "Unable to identify deployed instances", :red)
            exit(1)
          end # begin ... rescue
        end # def list

        desc "status ENVIRONMENT_INSTANCE_IDENTIFIER", "Determine the status of a deployed environment instance"
        def status (environment_instance_id = nil)
          raise Thor::Error.new(%Q{"status" was called incorrectly. Call as "status ENVIRONMENT_INSTANCE_IDENTIFIER"}) unless environment_instance_id
          begin
            create_coordinator()
            coordinator = Weaver::Deployers::Execution::CoordinationUtil::get_coordinator
            tx = coordinator.join_transaction(environment_instance_id)
#            tx = Weaver::Coordination::CoordinatedTransactionManager.instance.join_transaction(environment_instance_id, coordinator)
            @@logger.debug "joined status transaction #{tx.inspect}"
          rescue Exception => e
            # coordinator not available
            @@logger.warn e.inspect
            tx = nil
          end
          begin
            env = instance_model(environment_instance_id)
  
            # Print results
            table = []          
            table << ['Node ID', 'Hostname', 'IP Address', 'VM Status', 'Config Status']
            table << ['-------', '--------', '----------', '---------', '-------------']
            say_status("Done", "Summary of nodes:", :green)
            say() # empty line
            env.flatten_children.select { |n| n.respond_to?(:realized_by) && n.realized_by.empty? && !n.tags.member?(:storage) && !Weaver::Transforms::Unroll.template?(n)}.each do |node|
              # read hostname and ip address from model
              hostname = node.hostname || '-'
              ip_address = node.ip_address || '-'
              # read status from model if available (otherwise set to default)
              vm_status = Weaver::Coordination::VM_STATUS::UNKNOWN
              vm_status = node.amd[:vm_status].value unless node.amd[:vm_status].nil?
              config_status = Weaver::Coordination::VM_STATUS::UNKNOWN
              config_status = node.amd[:config_status].value unless node.amd[:config_status].nil?
              # read updated status from coordinator if available
              unless tx.nil?
                vm_status = tx.read_data("#{node.fully_qualified_id}.vm_status", :wait => false) || vm_status
                config_status = tx.read_data("#{node.fully_qualified_id}.config_status", :wait => false) || config_status
              end # unless tx.nil?
              table << ["#{node.id}", hostname, ip_address, vm_status, config_status]
            end # status.each
            print_table(table, {:indent => 14})
          rescue Exception => e
            @@logger.warn e.inspect
            say_status("Error", "Unable to read status for instance: #{environment_instance_id}", :red)
            exit(1)
          end # begin ... rescue
        end # method status

        desc "delete ENVIRONMENT_INSTANCE_IDENTIFIER", "Delete a deployed environment instance"
        method_option :config, :required => false, :type => :hash, :aliases => "-c", :desc => "Configuration parameters.", :default => {}
        method_option :deploy_params, :required => false, :type => :string, :aliases => "-d", :desc => "Deployment parameters.  DEPRECATED"
        
        def delete (environment_instance_id = nil)
          raise Thor::Error.new(%Q{"delete" was called incorrectly. Call as "delete ENVIRONMENT_INSTANCE_IDENTIFIER"}) unless environment_instance_id
          
          loader = init(options[:config])

          deploy_params = loader.deployment_parameters(parse_deploy_parameters(options[:deploy_params]))
          properties = loader.config_params.merge({ :log_level => @@logger.level, :timeout => options[:timeout]})

          begin
            create_coordinator()
            deployed_model = "#{DEPLOYMENT_DIR}/#{environment_instance_id}/model.weaver"
            env = Environment.new(deployed_model)
            # env = instance_model(environment_instance_id)
            env.deploy_or_update(:delete_deployment, {}, environment_instance_id, properties, deploy_params)
            say_status("Deleted", "#{environment_instance_id}", :green)
          rescue Exception => e
            @@logger.warn e.inspect
            say_status("Error", "Unable to delete instance: #{environment_instance_id}", :red)
            exit(1)
          end # begin ... rescue
        end # method status

        # update should not be available in the INTEGRATION stream
        desc "update INSTANCE_ID TEMPLATE ENVIRONMENT_FILE", "Update a deployed environment."
        method_option :template_config, :required => false, :type => :string, :alias => "-t", :desc => "Update properties"
        method_option :config, :required => false, :type => :hash, :aliases => "-c", :desc => "Configuration parameters.", :default => {}
        method_option :deploy_params, :required => false, :type => :string, :aliases => "-d", :desc => "Deployment parameters. DEPRECATED"
        method_option :timeout, :type => :numeric, :aliases => "-o", :desc => "Timeout for the deployment request (in seconds)"

        # Update an IAAS Pattern modeled as a Weaver file
        def update(instance_id, template, environment_file = nil)
          # This is a bug in Thor and JRuby so we have to handle that manually
          raise Thor::Error.new(%Q{"update" was called incorrectly. Call as "update INSTANCE_ID TEMPLATE ENVIRONMENT_FILE".}) unless instance_id
          raise Thor::Error.new(%Q{"update" was called incorrectly. Call as "update INSTANCE_ID TEMPLATE ENVIRONMENT_FILE".}) unless template
          
          loader = init(options[:config])
          raise WeaverError.new("Update not supported.") unless use_templates

          # load environment from config if not given as CLI argument
          if environment_file.nil?
            environment_file = loader.default_environment
          end
          validate_file_exist?(environment_file)

          # TODO hack to fix w/ service
          validate_file_exist?("/tmp/my_instances/#{instance_id}/model.weaver")
          say_status("Updating", "Environment '#{environment_file}'", :green)

          deploy_params = loader.deployment_parameters(options[:deploy_params])
          properties = loader.config_params.merge({ :log_level => @@logger.level, :timeout => options[:timeout]})
            
          template_config = Hash[*options[:template_config].split.map {|e|e.split('=')}.flatten] unless options[:template_config].nil?

          begin
            @@logger.debug("Update parameters: #{deploy_params}")
            e =  Weaver::Environment.new(environment_file, loader) { |lvl, status, msg| weaver_report(lvl, status, msg) }
            e.deploy_or_update(template, template_config, instance_id, properties, deploy_params)
            
            # Output results of command execution
            if e.root.topologies.values.first.amd[:deployment]
              say_status("Deployment id", "#{e.root.topologies.values.first.amd[:deployment].value} (#{e.root.topologies.values.first.amd[:config_status].value})", :green)
            end # if e.root.topologies.values.first.amd[:deployment]

            say_status("Done", "Summary of new or updated nodes:", :green)
            table = []
            table << ['Node ID', 'Hostname', 'IP Address', 'Status']
            table << ['-------', '--------', '----------', '-------']
            # TODO this needs to be simplified
            e.topology.flatten_children.select { |n| n.respond_to?(:realized_by) && n.realized_by.empty? && !n.tags.member?(:storage) && !Weaver::Transforms::Unroll.template?(n)}.each do |v|
              config_status = v.amd[:config_status].nil? ? "unknown" : "#{v.amd[:config_status].value}"
              #say_status("Node '#{v.id.to_s}'", "#{v.hostname} - #{v.ip_address}   #{config_status}", :green)
              table << ["#{v.id.to_s}", "#{v.hostname}", "#{v.ip_address}", "#{config_status}"]
            end
            say() # empty line
            print_table(table, {:indent => 14})

          rescue Timeout::Error => te
            @@logger.warn "Fatal error #{te.inspect} parsing '#{environment_file}'; backtrace: #{te.backtrace.join("\n")}"
            say_status("Error", "Timeout exceeded. Please use the '--timeout' option to increase the deployment timeout.",  :red)
            exit(1)
          rescue WeaverError => ex
            @@logger.warn "Fatal error #{ex.inspect} parsing '#{environment_file}'; backtrace: #{ex.backtrace.join("\n")}"
            if ex.original.kind_of?(Timeout::Error) # for some reason Timeout:Error get wrapped into WeaverError in 1.9
              say_status("Error", "Timeout exceeded. Please use the '--timeout' option to increase the deployment timeout.",  :red)
            else
              say_status("Error", ex.message, :red)
            end
            exit(1)
          rescue RuntimeError => re
            @@logger.warn "Fatal error #{re.inspect} parsing '#{environment_file}'; backtrace: #{re.backtrace.join("\n")}"
            say_status("Error", re.to_s, :red)
            exit(1)
          rescue Exception => e
            @@logger.warn "Fatal error #{e.inspect} parsing '#{environment_file}'; backtrace: #{e.backtrace.join("\n")}"
            say_status("Error", "Failed to deploy #{e.inspect}",  :red)
            exit(1)
          end
        end

        desc "generate ENVIRONMENT_ID", "Create an environment template '<ENVIRONMENT_ID>.weaver'"
        method_option :name, :type => :string, :aliases => "-n", :required => false, :desc => "Environment name"
        method_option :description, :type => :string, :aliases => "-d", :required => false, :desc => "Environment description"
        method_option :inf, :type => :string, :aliases => "-i", :required => true, :desc => "Infrastructure definition file"
        method_option :app, :type => :string, :aliases => "-a", :required => false, :desc => "Application definition file"
        method_option :bindings, :type => :hash, :aliases => "-b", :required => false, :desc => "Application to infrastructure node bindings"
        method_option :no_defaults, :type => :boolean, :aliases => "-d", :default => true, :required => false, :desc => "Inherit default values from inf and app definitions"
        method_option :loc, :type => :string, :aliases => "-l", :desc => "Location of the file."

        # Create a .weaver file including an infrastructure and optional application Topology.
        def generate(environment_id = nil)
          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"generate" was called incorrectly. Call as "generate ENVIRONMENT_ID".}) unless environment_id
          collect_properties
          setup_logging
          validate_file_exist?(options[:app]) if options[:app] # is optional
          validate_file_exist?(options[:inf])

          filename = environment_id.to_s + ".weaver"
          if options[:loc]
            filename = File.join(options[:loc], filename)
          end
          say_status("Creating:", "Environment definition '#{filename}'...", :green)

          generator = Weaver::Generators::EnvironmentGenerator.new(options[:app], options[:inf])

          # convert realizations Hash to symbols
          bindings = Hash[options[:bindings].map{ |k, v| [k.to_sym, v.to_sym] }] if options[:bindings]

          FileUtils.mkdir_p(File.dirname(filename)) unless File.directory?(File.dirname(filename))

          # creating a backup copy if file already exists
          if File.exists?(filename)
            say_status("Creating", "File '#{filename}' already exists. Creating backup '#{filename}.saved'", :green)
            FileUtils.copy(filename, "#{filename}.saved")
          end

          begin
            generator.generate_environment_skeleton(filename,
            environment_id, options[:name], options[:description], bindings, options[:no_defaults])
          rescue WeaverError => ex
            say_status("Error:", ex.message, :red)
            exit(1)
          end
        end
        
        desc "analytics [ENVIRONMENT_FILE]", "Collect and store data on the environment to support runtime compliance checking and problem determination."
        method_option :config, :required => false, :type => :hash, :aliases => "-c", :desc => "Configuration parameters.", :default => {}
        method_option :deploy_params, :required => false, :type => :string, :aliases => "-d", :desc => "Deployment parameters. DEPRECATED"
        method_option :timeout, :type => :numeric, :aliases => "-o", :desc => "Timeout for the deployment request (in seconds)"
        
        def analytics(environment_file = nil)
          loader = init(options[:config])
         
          # load environment from config if not given as CLI argument
          if environment_file.nil?
            environment_file = loader.default_environment
          end
          validate_file_exist?(environment_file)
         
          deploy_params = loader.deployment_parameters(parse_deploy_parameters(options[:deploy_params]))
          properties = loader.config_params.merge({ :log_level => @@logger.level, :timeout => options[:timeout]})
            
          if !use_service
             say_status("Error", "Please, specify the hostname and port of the Weaver service using system properties. Example: weaver env analytics --sysproperties=service_host:localhost service_port:9999", :red)
             exit(1)
          end
            
          say_status("Analyzing", "Environment '#{environment_file}'", :green)
            
          # Deploys the environment, using the service. The bootstrap parameter is set to true
          #  to trigger Weaver to learn continuous validation rules
          # and produce the "xprints" (fingerprints) of each piece of automation
          deploy_with_service(environment_file, properties, deploy_params, true)          
        end

        no_tasks do
          # Output a message to standard output
          def weaver_report(lvl, status, msg)

            if !options[:no_color]
              colors = { Logger::INFO => :green, Logger::WARN => :yellow, Logger::ERROR => :red, Logger::FATAL => :red, Logger::DEBUG => :clear }
            else
              colors = { }
            end

            # Ignore DEBUG messages, we don't want to actually display them, only :abort them on request from a GUI
            say_status(status, msg, colors[lvl]) unless lvl == Logger::DEBUG
          end

          # The full path of all files matching +filespec+.
          def files(filespec)
            return Dir.glob(filespec).map { |filename|
              File.new(File::expand_path(filename))
            }
          end
          
          def instance_model (instance_id)
            deployed_env = nil
            deployed_model = instance_id.nil? ? nil : "#{DEPLOYMENT_DIR}/#{instance_id}/model.weaver"
            if !instance_id.nil?
              raise WeaverError.new("Deployed instance model cannot be found") unless File.exists?(deployed_model)
              deployed_env = Weaver::Dsl::WeaverParser.new.parse(deployed_model)
              raise WeaverError.new("Deployed instance model invalid") if deployed_env.nil?
            end
            return deployed_env
          end # end instance_model
          
          def use_templates
            use = Weaver::System.instance.get_property("use_templates")
            use = false if use.nil?
            use = use.to_b if use.is_a?(String)
            return use
          end # def use_templates
                
          def use_service
            return(Weaver::System.instance.get_property("service_host") && !(Weaver::System.instance.get_property("service_host").empty?))
          end

          def deploy_with_service(environment_file, properties, deployment_parameters, analytics = false)
            begin
              say_status("Deploying", "Environment '#{environment_file}' with the Weaver Service", :green)
              # TODO: collect environment_variables - design how env vars will be passed to the service
              environment_variables = {}
              host = Weaver::System.instance.get_property("service_host")
              port = Weaver::System.instance.get_property("service_port")
              service_client = Weaver::Service::ServiceClient.new(host, port)
              
              # merge all system properties into 'properties[:system_settings]'
              (properties[:system_settings] = {}) if properties[:system_settings].nil?
              properties[:system_settings] = properties[:system_settings].merge(Weaver::System.instance.get_properties_as_string_hash.symbolize)
              
              # Indicates whether or not this deployment will be done for bootstrapping (analytics) purposes
              properties[:system_settings][:analytics] = analytics
            
              env = Environment.new(environment_file)
              id = service_client.post_deployment(env.root, properties, deployment_parameters, environment_variables)  
              @@logger.info "Successfully posted deployment id #{id}"
              if analytics
                say_status("Learning in progress...", "Deployment id '#{id}'", :green)
              else
                say_status("Deployment successfully posted", "Deployment id '#{id}'", :green)
              end
            rescue Timeout::Error => te
              @@logger.warn "Fatal error #{te.inspect} deploying '#{environment_file}'; backtrace: #{te.backtrace.join("\n")}"
              say_status("Error", "Timeout exceeded. Please use the '--timeout' option to increase the deployment timeout.",  :red)
              exit(1)
            rescue Weaver::Service::WeaverServiceError => e
              @@logger.warn "Fatal error #{e.inspect} deploying '#{environment_file}'; backtrace: #{e.backtrace.join("\n")}"
              say_status("Error", e.message, :red)
              exit(1)
            rescue WeaverError => e
              @@logger.warn "Fatal error #{e.inspect} deploying '#{environment_file}'; backtrace: #{e.backtrace.join("\n")}"
              if e.original.kind_of?(Timeout::Error) # for some reason Timeout:Error get wrapped into WeaverError in 1.9
                say_status("Error", "Timeout exceeded. Please use the '--timeout' option to increase the deployment timeout.",  :red)
              else
                say_status("Error", ex.message, :red)
              end
              exit(1)
            rescue => e
              @@logger.warn "Fatal error #{e.inspect} deploying '#{environment_file}'; backtrace: #{e.backtrace.join("\n")}"
              say_status("Error", e.message, :red)
              exit(1)
            end
          end
        end
      end
    end

  end
end
