require 'weaver/core/base'
require 'weaver/core/generation'
require 'weaver/core/automation_container'
require 'weaver/core/component_container_mixin'
require 'weaver/core/policy_container_mixin'
require 'weaver/core/import_container_mixin'
require 'weaver/core/topology_container_mixin'
require 'weaver/core/provider_container_mixin'
require 'weaver/core/node_container_mixin'
require 'weaver/core/user_container_mixin'
require 'weaver/core/group_container_mixin'
require 'weaver/core/link_container_mixin'

module Weaver
  module Core
    # Represents a 'provider', such as an IWD connection and credentials
    #
    # @see Weaver::Dsl::ProviderStatement
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    class Provider < Base
      include Weaver::Core::Generation
      
      # A symbol identifying the cloud 'provider', e.g. :iwd_vsp or :openstack
      attr_reader :provider_type
      
      # * +id+ A symbol for the Provider, e.g. +:amazon_ec2+
      # * +parent+ The enclosing Base
      def initialize(id, parent)
        super(id, parent)
        
        @type = :provider
        @mixins = {}
        begin
          # We special case :provider, which is not associated with an actual cloud provider model,
          # but is used as a model provider by Weaver's type system
          set_provider_type(id) unless id == :provider
        rescue Errno::ENOENT
          # Do nothing
       end
      end

      # Mix provider-specific attributes
      #  - +entity+ Core::Base
      def mixin_provider_attributes(entity)
        if @mixins.has_key?(entity.class)
          mixin = @mixins[entity.class]
        else
          # Look for Weaver type, if not found look at base Weaver types until a mixin is found
          mixin = nil
          candidate = entity
          begin
            mixin = @mixins[candidate.type.to_s]
            candidate = candidate.provenance
          end while mixin.nil? && !candidate.nil?
        end
        
        @@logger.debug "Enhancing #{entity.fully_qualified_id} using #{self.fully_qualified_id} and model #{mixin.inspect}"
           
        # Don't overwrite attributes that already exist
        unless mixin.nil?
          mixin.amd.each do |id, attr| 
            unless entity.amd.has_key?(id)
              entity.amd[id] = attr.clone
              entity.amd[id].scope = entity
              entity.amd[id].weaver_origin = entity.weaver_origin
            end 
          end
            
          unless mixin.mixer.nil?
            mixin.mixer.call(entity)
          end
          
          entity.provider_mixins << mixin.class
        else
          @@logger.warn "Cannot mix #{entity.type} with provider #{fully_qualified_id}"
#          raise WeaverError.new("Cannot mix #{entity.type} with provider #{fully_qualified_id}")
        end
      end
      
      # Returns a Weaver::Deployers::Deployer      
      def deployer(properties = {}, &reporter)
        provider()
        raise WeaverError.new("#{fully_qualified_id} is not associated with a deployer") if @provider_impl.nil?
        return @provider_impl.deployer(properties, &reporter)
      end
      
      # Load type definition
      #  - +typeid+ provider type, e.g. :amazon_ec2 or :iwd_vsp
      def set_provider_type(typeid)
        @provider_type = typeid
        
        # Every Provider is associated with a Weaver file that contains its mixins
        mixin_files = Gem.find_files("weaver/providers/#{typeid.to_s}.weaver")
        if mixin_files.empty?
          @@logger.warn "Could not find 'weaver/providers/#{typeid.to_s}.weaver'"
        else
          @library = Weaver::Dsl::WeaverParser.new().parse(mixin_files[0])
          
          # Augment ourselves
          provider_mixin = @library.prototype("#{@provider_type}_provider_mixin".to_sym)
          provider_mixin.amd.each do |id, attr|
            attr = attr.clone
            attr.scope = self
            attr.weaver_origin = weaver_origin 
            @amd[id] = attr
          end
            
          # Save all types ending in _mixin
          @mixins = {}
          @library.weaver_extended_types.each do |typeid, entity|
            md = typeid.to_s.match(Regexp.new("^#{provider_type}_([a-z_]+)_mixin$"))
            unless md.nil?
              mixin = @library.prototype(typeid.to_sym)
              @mixins[md[1]] = mixin
            end
          end
          
#          # Look up mixins for Topology and Node for use when we find those attributes
#          [ "#{@provider_type}_node_mixin", "#{@provider_type}_topology_mixin"].each do |typeid|
#            mixin = @library.prototype(typeid.to_sym)
#            type = root_entity.prototype(typeid.split('_')[-2].to_sym).class
#            @mixins[type] = mixin
#          end
        end
        
      end

      def validate_collect(out_messages, desired_phases = nil)
        super

        if @library.nil?
          report(Logger::WARN, "Unknown provider type #{@provider_type.inspect}", self)
        end      
      end
      
      # Validates that +entity+ can be published by this provider.  An example
      # is that for example only some providers handle elastic_ip
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_publishability(entity, out_messages, desired_phases)
        @provider_impl.validate_publishability(entity, out_messages, desired_phases) unless provider.nil? 
      end

      # Returns :artifact or :cloud depending on the kind of Provider
      def provider_class
        return amd.key?(:provider_class) ? amd[:provider_class].value : :cloud
      end

      protected
      
      def provider
        if @provider_impl.nil?
          begin
            provider_impl_class = Kernel.eval("Weaver::Providers::#{@provider_type.to_s.capitalize}")
            @provider_impl = provider_impl_class.new(self)
          rescue LoadError
            puts "No Ruby implementation for provider #{@provider_type}"
            @@logger.warn "No Ruby implementation for provider #{@provider_type}"
          rescue NameError
            puts "NameError for provider #{@provider_type}; no class Weaver::Providers::#{@provider_type.to_s.capitalize}"
            @@logger.warn "NameError for provider #{@provider_type}; no class Weaver::Providers::#{@provider_type.to_s.capitalize}"
          rescue Exception => ex
            puts "#{ex} creating Weaver::Providers::#{@provider_type.to_s.capitalize} for provider #{@provider_type}"
            @@logger.warn "#{ex} creating Weaver::Providers::#{@provider_type.to_s.capitalize} for provider #{@provider_type}"
          end
        end
        
        return @provider_impl
      end
              
    end
  end
end