module Weaver
  module Core
    # Represent a requirement place by a Component upon infrastructure
    #
    # Author:: Ed Snible
    class Requirement
      include Weaver::Core::Generation
      include Weaver::Validation

      # Where (in a .weaver source file) was this entity defined
      attr_accessor :weaver_origin
      
      # The Component this Requirement is attached to
      attr_accessor :scope
      
      # * +scope+ This requirement's container, typically a Component
      def initialize(scope)
        @scope = scope
        raise "Scope must be a Base" unless scope.kind_of?(Base)
      end
      
      # RECURSIVE
      # * +node+ a Node
      # * +result_accum+ Set of bound Nodes that directly or indirectly realize a Node
      # * +intermediate_accum+ Set of unbound Nodes that directly or indirectly realize a Node; the set includes the Node itself
      def all_final_realizations_accum(node, result_accum, intermediate_accum)
        # Is +node+ bound?
        if node.realized_by.empty?
          result_accum.add(node)
        else
          raise WeaverError.new("realized_by cycle: #{result_accum.to_a.map { |entity| entity.fully_qualified_id }.inspect}") if intermediate_accum.member?(node)
          intermediate_accum.add(node)
          
          # We ignore errors in realizations, such as Symbol instead of Node, expecting other validation to catch that
          node.realized_by.each do |realization|
            if realization.kind_of?(Node)
              all_final_realizations_accum(realization, result_accum, intermediate_accum)
            else
              # If the node is something bogus, like a Symbol, then it is a result as well (as an intermediate)
              result_accum << node
            end
          end
        end
      end
      
      # Find all bound Nodes that are realizations of the owning Node
      def all_final_realizations()
        node = get_owning_node()
        
        if node.realized_by.empty?
          # We evaluate requirements on infrastructure nodes against themselves.  
          return [node] unless node.cloud.nil?
          
          # We ignore unrealized app nodes
          return []
        end
        
        accum = Set.new
        all_final_realizations_accum(node, accum, Set.new)
        return accum.to_a
      end
      
    protected
    
    # The nearest Node ancestor walking up the Weaver model to the root
    def get_owning_node()
      ancestor = @scope
      while !ancestor.nil? && !ancestor.kind_of?(Node)
        ancestor = ancestor.scope
      end
      return ancestor
    end
    
      def get_infrastructure_value(base, key)
#        @@logger.debug "key is #{key}"
        @@logger.debug "base.#{key} is #{base.amd.has_key?(key) ? base.amd[key].value.inspect : nil}"
#        @@logger.debug "base.amd.keys is #{base.amd.keys.inspect}"
        inf_value = base.amd.has_key?(key) ? base.amd[key].value : nil
        if inf_value.nil? && base.amd.keys.member?(key)
          inf_value = :NO_DEFAULT
        end
        @@logger.debug "inf_value is #{inf_value}"
        return inf_value
       end
       
      # TODO Refactor this into a core utility
      def weaver_classname(base)
         case
         when base.kind_of?(Component)
           return "component"
         when base.kind_of?(Node)
           return "node"
         else
           raise "Unexpected Weaver Type #{base.class}"
         end
      end
       
      # Return a String containing a validation message
      def generate_undefined_message(target, value)
        return <<-eos
#{weaver_classname(@scope)} :#{@scope.id} at (#{@weaver_origin}) requires the
#{weaver_classname(target)} :#{target.id} at (#{target.weaver_origin}) to have
 
  attribute :#{@key} => #{value}
  
Configure the cloud image to meet this requirement and rediscover.  If discovery did not detect the
requirement but you are certain it is there you may add the attribute to the infrastructure node.

eos
      end
      
      # Returns true if this Requirement is equal to another Requirement
      # TODO write these for subclasses?
      def ==(other)
        # We ignore scope, and compare based only on content not where we are placed
        return true
      end
      
    end
    
    # Represents a Requirement that can be checked using set membership 
    #
    # Author:: Ed Snible
    class MemberRequirement < Requirement
      # * +scope+ This requirement's container, typically a Component
      # * +h+ Hash of requirement attributes
      def initialize(scope, h)
        super scope
        
        h = h.clone()
        @@logger.warn ":version in requirement not supported" if h.include?(:version)
        h.delete(:version)
        h.delete(:description)
        
        raise WeaverError.new("MemberRequirement has multiple constraints: #{h}") if h.size > 1
        
        @key = h.keys[0]
        @value = h.values[0]
      end
      
      # Validate this requirement
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be <code>[:predefined]</code>
      def validate_collect(out_messages, desired_phases = nil)
        # TODO refactor so that only the line "!inf_value.member?(@value)" needs to be written
#        inf_node = get_final_realization()
        all_final_realizations.each do |inf_node|
          inf_value = get_infrastructure_value(inf_node, @key)
          if inf_value.nil?
            out_messages.push ProblemStatus.new(Logger::ERROR, generate_undefined_message(inf_node, "['#{@value}']"), inf_node, weaver_origin)
          elsif inf_value.kind_of?(Array)
            if !inf_value.member?(@value)
              out_messages.push ProblemStatus.new(Logger::ERROR, "Node :#{inf_node.id} does not have :#{@key} => ['#{@value}', ...]", inf_node, weaver_origin)
            end
          else
            out_messages.push ProblemStatus.new(Logger::ERROR, "Expected #{inf_node.id.inspect} to define { #{@key.inspect} => [#{@value.inspect}, ...] } but #{inf_node.id.inspect} has { #{@key.inspect} => #{inf_value.inspect} }", inf_node, weaver_origin) 
          end
        end
      end
      
    end
    
    # Represents a Requirement that can be checked using subset checking 
    #
    # Author:: Ed Snible
    class SubsetRequirement < Requirement
      # * +scope+ This requirement's container, typically a Component
      # * +h+ Hash of requirement attributes
      def initialize(scope, h)
        super scope
        
        raise WeaverError.new("SubsetRequirement has multiple constraints: #{h}") if h.size > 1
        
        @key = h.keys[0]
        @subset = Set.new(h.values[0])
      end
      
      # Validate this requirement
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be <code>[:predefined]</code>
      def validate_collect(out_messages, desired_phases = nil)
        # TODO refactor so that only the line "!inf_value.member?(@value)" needs to be written
#        inf_node = get_final_realization()
        all_final_realizations.each do |inf_node|
          inf_value_arr = get_infrastructure_value(inf_node, @key)
          if inf_value_arr.nil?
            out_messages.push ProblemStatus.new(Logger::ERROR, "Cannot verify #{self.class}, attributes[:#{@key}] undefined on #{inf_node.id}", inf_node, weaver_origin)
          elsif inf_value_arr.kind_of?(Array)
            as_set = Set.new(inf_value_arr)
            if !@subset.subset?(as_set)
              missing_values = @subset.clone
              missing_values.subtract(as_set)
              out_messages.push ProblemStatus.new(Logger::ERROR, "node :#{inf_node.id} missing 'attribute[:#{@key}] => #{missing_values.to_a}'", inf_node, weaver_origin)
            end
          else
            out_messages.push ProblemStatus.new(Logger::ERROR, "Expected node :#{inf_node.id} { #{@key} => #{@subset.to_a} } but found type was #{inf_value_arr.class}", inf_node, weaver_origin) 
          end
        end
      end
    end
    
    # Represents a Requirement that can be checked using set membership 
    #
    # Author:: Ed Snible
    class NonmemberRequirement < Requirement
      # * +scope+ This requirement's container, typically a Component
      # * +h+ Hash of requirement attributes
      def initialize(scope, h)
        super scope
        
        raise WeaverError.new("NonmemberRequirement has multiple constraints: #{h}") if h.size > 1
        
        @key = h.keys[0]
        @value = h.values[0]
      end
      
      # Validate this requirement
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be <code>[:predefined]</code>
      def validate_collect(out_messages, desired_phases = nil)
        # TODO refactor so that only the line "!inf_value.member?(@value)" needs to be written
        all_final_realizations.each do |inf_node|
          inf_value_set = get_infrastructure_value(inf_node, @key)
          if inf_value_set.nil?
            out_messages.push ProblemStatus.new(Logger::ERROR, "Cannot verify #{self.class}, #{@key} undefined on #{inf_node.id}", inf_node, weaver_origin)
          elsif inf_value_set.kind_of?(Array)
            if inf_value_set.member?(@value)
              out_messages.push ProblemStatus.new(Logger::ERROR, "#{inf_node.id} already using #{@key} => #{@value}", inf_node, weaver_origin)
            end
          else
            out_messages.push ProblemStatus.new(Logger::ERROR, "Expected node :#{inf_node.id} { #{@key} => ['#{@value}', ...] } but found type was #{inf_value_set.class}", inf_node, weaver_origin) 
          end
        end
      end
    end

    # Represents a Requirement that can be checked using set intersection 
    #
    # Author:: Ed Snible
    class EmptyIntersectionRequirement < Requirement
      # * +scope+ This requirement's container, typically a Component
      # * +h+ Hash of requirement attributes
      def initialize(scope, h)
        super scope
        
        raise WeaverError.new("EmptyIntersectionRequirement has multiple constraints: #{h}") if h.size > 1
        
        @key = h.keys[0]
        @set = Set.new(h.values[0])
      end
      
      # Validate this requirement
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be <code>[:predefined]</code>
      def validate_collect(out_messages, desired_phases = nil)
        # TODO refactor so that only the line "!inf_value.member?(@value)" needs to be written
#        inf_node = get_final_realization()
        all_final_realizations.each do |inf_node|
          inf_value_arr = get_infrastructure_value(inf_node, @key)
          if inf_value_arr.nil?
            out_messages.push ProblemStatus.new(Logger::ERROR, generate_undefined_message(inf_node, @set.to_a), inf_node, weaver_origin)
          elsif inf_value_arr.kind_of?(Array)
            intersection = @set.intersection(inf_value_arr)
            if intersection.size > 0
              out_messages.push ProblemStatus.new(Logger::ERROR, "node #{inf_node.id} { attributes :#{@key} } already contains #{intersection.to_a}", inf_node, weaver_origin)
              out_messages.push ProblemStatus.new(Logger::ERROR, generate_mismatch_message(inf_node, intersection.to_a), weaver_origin)
            end
          else
            out_messages.push ProblemStatus.new(Logger::ERROR, "Expected node :#{inf_node.id} { #{@key} => #{@set.to_a} } but found type was #{inf_value_arr.class}", inf_node, weaver_origin) 
          end
        end
      end
      
      # Return a String containing a validation message
      def generate_mismatch_message(target, intersection)
        return <<-eos
#{weaver_classname(@scope)} :#{@scope.id} at (#{@weaver_origin}) requires the
#{weaver_classname(target)} :#{target.id} at (#{target.weaver_origin}) to not already be using
 
  attribute :#{@key} => #{intersection.to_a}

eos
      end
    end
    
  end
end