require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Topology which may contain Topologies
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module TopologyContainerMixin
      # Let all classes that include TopologyContainerMixin have a method to return their topologies
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Topology
          def topologies
            # Since mixins have no initialize() to set up @topologies, initialize @topologies lazily
            @topologies ||= {} unless frozen?
            return @topologies || {}
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_topologies)
        baseclazz.register_validate_handler(:validate_topologies)
        baseclazz.register_child_enumerator(:topologies)           
        
      end
      
      # Deeply copy @topologies
      def initialize_copy_topologies(other_topology_container)
      
        # At this point, self.initialize_copy should have cloned @topologies but the values are shallow.
        @topologies = {}
        other_topology_container.topologies.each do |id, topology|
          raise WeaverError.new("topology (#{id.inspect}) is nil") if topology.nil?
          dup = topology.clone()
          dup.scope = self
          @topologies[id] = dup
        end
      end

      # Validate @topologies      
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_topologies(out_messages, desired_phases = nil)
        topologies.each { |id, topology| topology.validate_collect(out_messages, desired_phases) }
      end
    end
    
  end
end