require 'weaver/core/rule'
require 'weaver/core/resolution'
require 'weaver/core/continuous_validation'
require 'weaver/dsl/resolution_statement'

module Weaver
  module Dsl
    # The RuleStatement implements the processing of the language element 'rule'.
    #
    # Author:: Ed Snible
    #
    class RuleStatement
      include Weaver::Core

      # The model represents the Core::Rule that corresponds to this statement
      attr_accessor :model
      
      # Represents the enclosing statement
      attr_accessor :parent
      
      # * +id+ Symbol representing the Core::Rule
      # * +parent_statement+ The enclosing AbstractStatement
      def initialize(id, parent_statement)
        @@logger.debug "rule[id: #{id}]"
        @model = Rule.new(id)
        @parent = parent_statement
      end
      
      # Parse the 'rule_id' of a Weaver rule
      def rule_id(rule_id)
        @@logger.debug "rule_id = #{rule_id}"
        @model.rule_id = rule_id
      end
      
      # Parse the 'description' of a Weaver rule
      def description(rule_description)
        @@logger.debug "description = #{rule_description}"
        @model.description = rule_description
      end

      # Parse the 'phase' of a Weaver rule
      def phase(validation_phase = nil, &block)
        if !block.nil?
          # Make sure we do not parse the notation { :phase }
          raise WeaverError.new("Rule phase must be a valid symbol: :predeployment or :postdeployment or :continuous") unless (validation_phase == :predeployment || validation_phase == :postdeployment || validation_phase == :continuous)
        end
        unless validation_phase.nil?
          raise WeaverError.new("Rule phase must be a valid symbol: :predeployment or :postdeployment or :continuous") unless (validation_phase == :predeployment || validation_phase == :postdeployment || validation_phase == :continuous)
        end
        @@logger.debug "phase[:#{validation_phase}]"
        @model.phase = validation_phase
      end
      
      # Parse the 'tags' of a Weaver rule
      def tags(*tag_array)
        @@logger.debug "tags = #{tag_array.inspect}"
#       @model.tags = *tag_array      (Fixed: In 1.6.7 *tag_array with 1 element returns the element, not an array reference)   
        @model.tags = []
        unless tag_array.nil?
          tag_array.each do |tag|
            @model.tags << tag
          end
        end
      end           
      
      # Process +resolution+ keyword.
      #
      # @see ResolutionStatement
      def resolution(&block)
        resolution_statement = ResolutionStatement.new(self)
        resolution_statement.instance_eval(&block)
        resolution = resolution_statement.model
        add_resolution(resolution.label, resolution.description, resolution.type, resolution.fix_proc) unless resolution.nil?
      end

      # Parse the 'where' clause of a Weaver rule
      def where(&block)
        @@logger.debug "where[]"
        raise WeaverError.new("Error in rule :#{@model.id}. Multiple 'where' clauses in rule") unless @model.where.nil?
        @model.where = block
      end
      
      # Parse the 'body' of a Weaver rule
      def body(&block)
        @@logger.debug "body[]"
        raise WeaverError.new("Error in rule :#{@model.id}. Multiple 'body' in rule") unless @model.body.nil?
        raise WeaverError.new("Error in rule :#{@model.id}. A rule with a 'body' cannot have a 'continuous_validation' block.") unless @model.contval.nil?
        @model.body = lambda &block
      end
      
      # Provide a way for the body to call report()
      # TODO Replace this with running rules within the rule, not the statement-of-origin
      def report(*args)
        return @model.report(*args)
      end
      
      # Provide a way for the body to call add_resolution()
      # TODO Replace this with running rules within the rule, not the statement-of-origin
      def add_resolution(*args)
        return @model.add_resolution(*args)
      end
      
      # Parse the 'continuous_validation' block of a rule
      # * +block+ -> Block containing a Weaver declaration describing the continuous validation elements
      def continuous_validation(&block)
        @@logger.debug "continuous_validation[]"
        raise WeaverError.new("Error in rule :#{@model.id}. Multiple 'continuous_validation' blocks.") unless @model.contval.nil?
        raise WeaverError.new("Error in rule :#{@model.id}. A continuous validation rule cannot have a body.") unless @model.body.nil?
        contval_statement = Weaver::Dsl::ContinuousValidationStatement.new(self)
        contval_statement.instance_eval &block
      end
    end
    
    # The ContinuousValidationStatement class implements the processing of the 'continuous_validation' block of a rule.
    # Author:: Fabio Oliveira
    class ContinuousValidationStatement
      include Weaver::Core
      # * +parent+ - The enclosing RuleStatement
      def initialize(parent)
        @parent = parent
        @parent.model.contval = Weaver::Core::ContinuousValidation.new(@parent.model)
      end
      
      # Parse the 'runtime' keyword of the 'continuous_validation' block of a rule. 
      # Following the 'runtime' keyword is a Hash supporting the following keys:
      # +:type+ - The runtime used to run the continuous validation rule. Valid value :cron
      # +:period+ - Integer corresponding to the running period of the continuous validation rule. The value is given in minutes.
      # If the period is 3 minutes, the rule will run every 3 minutes.
      # Example: runtime :type => cron, :period => 3
      def runtime(runtime_metadata)
        raise WeaverError.new("Error in rule :#{@parent.model.id}. Incorrect format of runtime inside a continuous validation block. Expected: 'runtime :type => <type>, :period => <period in minutes>'") unless runtime_metadata.kind_of?(Hash)
        @@logger.debug "continuous_validation: runtime_metadata = #{runtime_metadata.inspect}"
        @@logger.debug "continuous_validation: runtime type = #{runtime_metadata[:type]}"
        @@logger.debug "continuous_validation: period = #{runtime_metadata[:period]}"
        @parent.model.contval.runtime_type = runtime_metadata[:type]
        @parent.model.contval.period = runtime_metadata[:period]
      end
      
      # Parse the 'command' keyword of the 'continuous_validation' block of a rule.
      def command(command_str)
        raise WeaverError.new("Error in rule :#{@parent.model.id}. Inside a continuous_validation block, 'command' must be a String.") unless command_str.kind_of?(String)
        @@logger.debug "continuous validation: command = '#{command_str}'"
        @parent.model.contval.command = command_str
      end
      
      # Parse the 'artifact' keyword of the 'continuous_validation' block of a rule
      def artifact(*artifacts)
        @@logger.debug "continuous validation: artifacts = #{artifacts.inspect}"
        @parent.model.contval.artifacts = [] 
        artifacts.each do |a|
          @parent.model.contval.artifacts << a
        end
      end
      
      def method_missing(sym, *args, &block)
        @@logger.debug "Trying to resolve reference made by continuous validation rule #{@parent.model.id}: symbol = #{sym}; args = #{args.inspect}; block =  #{block.inspect}"
        if @parent.model.contval.artifacts.include?(sym)
          @@logger.debug "Resolved artifact reference: #{@parent.parent.model.root_entity.artifacts[sym].inspect}"
          return @parent.parent.model.root_entity.artifacts[sym]
        else
          msg = "Invalid reference made by continuous validation rule #{@parent.model.id}: symbol = #{sym}; args = #{args.inspect}; block =  #{block.inspect}" 
          raise WeaverError.new(msg)
        end
      end
      
      def late_binding(&block)
        block_contents = block.to_source({:strip_enclosure => true})
        ret = block_contents
        l = block_contents.length
        if block_contents[0] == '"' && block_contents[l - 1] == '"'
          @@logger.debug "Removed leading and trailing double-quote characters of #{block_contents}"
          ret = block_contents[1, l - 2]
        end
        @@logger.debug "Deferring late_binding creation in rule #{@parent.model.id} for block '#{block_contents}'"
        return ret
      end
    end
    
  end
end