require 'weaver/dsl/weaver_parser'
require 'weaver/chef/chef_node_inspector'

module Weaver
  module Generators
    # The PatternGenerator generates a Weaver {Weaver::Core::Topology} from a list of IaaS {Weaver::Core::Node Nodes}
    # and a mapping of node names.
    class PatternGenerator
      include Weaver::Core

      # Generates an infrastructure pattern made up of images.
      #
      # * *Args*    :
      #   - +pattern_name+ -> a String; the name of the pattern Infrastructure
      #   - +weaver_images+ -> Hash of name (a String) to Root
      #   - +image_mapping+ -> Hash of \Weaver node name (a String) to cloud image name-or-id (also a String)
      #   - +provisioner+ -> a Provisioner with credentials
      #
      # * *Returns* :
      #   - the generated Infrastructure model
      #
      def self.generate_infrastructure(pattern_name, weaver_images, image_mapping, provisioner)

        inf_root = Weaver::Dsl::WeaverParser.new().parse(nil)
        inf_root.id = pattern_name + Weaver::Core::WEAVER_SUFFIX
        
#        inf_root.preamble = "properties = load_properties(File.join(File.dirname(__FILE__), #{config_file.inspect}))"
        
        inf_root.prefix_comments << "Generated by weaver on #{DateTime.now.strftime("%FT%R")}" 
        inf_root.instance_exec {
          @raw_weaver_imports = {} if @raw_weaver_imports.nil?
          @raw_weaver_imports[:images] = weaver_images.values.map { |root| root.id }
        }
        
        inf_top = Topology.new(Weaver::Core::WeaverTransforms.rubyize_id(pattern_name.symbolize), inf_root)
        inf_root.topologies[inf_top.id] = inf_top
        inf_top.providers[provisioner.id] = provisioner

        basic_provider_attributes = inf_root.prototype(:provider).amd.keys
        provisioner.amd.each_value do |attr| 
          unless basic_provider_attributes.member?(attr.id)
            attr.value = Weaver::Chef::UnquotedString.new("config[#{provisioner.id.inspect}][#{attr.id.inspect}]")
            attr.modeled = true
          end
        end        
        
        image_mapping.each do |node_name, name_or_id|
          root = weaver_images[name_or_id]
          dummy_root = Weaver::Dsl::WeaverParser.new().parse(nil)
          dummy_import = Import.new(:images, dummy_root)
          dummy_node = Weaver::Transforms::Duplicate.duplicate(root.nodes.values[0])
          dummy_node.scope = dummy_import 
          dummy_import.nodes[dummy_node.id] = dummy_node
          inf_top.unresolved_refinements[node_name.symbolize] = dummy_node
          
          inf_top.unresolved_additional_declarations << "#{node_name.symbolize}.name = #{node_name.inspect}"
        end
        
        return inf_root
      end

    end

  end

end
