require 'erubis'
require 'string'
require 'json'
require 'weaver/core/logging'   
require 'weaver/core/weaver_error'

module Weaver
  module Iwd
    # Generates a IWD script package creation JSON request from a Weaver
    # infrastructure definition.
    #
    # Author:: Ann Marie Fred
    #
    class IwdScriptGenerator
      
      # Creates a new IwdScriptGenerator for a given Environment definition.
      #
      # * *Args*    :
      #   - +name+ -> name of script package
      #   - +version+ -> version of script package
      #   - +description+ -> description of script package
      #   - +role_file+ -> chef role file
      #   - +env_model+ -> Weaver environment definition file
      #
      # * *Returns* :
      #   - an IwdRequestGenerator instance
      #
      def initialize(script_attributes)
         @@logger.debug("Weaver::Iwd::IwdScriptGenerator.new starting")
         @@logger.debug(script_attributes)
        
         @name = script_attributes['name']
         @version = script_attributes['version']
         @description = script_attributes['description'] 
         if script_attributes['description'].nil?
         	@description = "Script package #{@name}_#{@version}"
         end
         @rolefile = script_attributes['role_file']
         @model = script_attributes['env_model']
         
         raise WeaverError.new("Name required to define script package") unless @name
         raise WeaverError.new("Version required to define script package") unless @version
         raise WeaverError.new("Role file required to define script package") unless @rolefile
         raise WeaverError.new("Model required to define script package") unless @model
                 
        @@logger.debug("Weaver::Iwd::IwdScriptGenerator.new ending")
      end # end initialize method

      # Generates a cbscript.json file for a script package.
      #
      # * *Returns* :
      #   - the JSON metadata for a script package
      def to_json()
        @@logger.debug("Weaver::Iwd::IwdScriptGenerator.to_json starting") 
        json = {
              'name' => @name,
              'version' => @version,
              'description' => @description,
              'command' => "/usr/bin/ruby",
              'log' => "/tmp/#{@name}",
              'location' => "/tmp/#{@name}",
              'timeout' => "0",
              'commandargs' => "/tmp/#{@name}/chef_script_run.sh"
               }
               
        @keys = add_args(@rolefile, @model)
               
        json['keys'] = @keys

        #puts JSON.pretty_generate(json)
        @@logger.debug("Weaver::Iwd::IwdScriptGenerator.to_json ending") 
        return JSON.pretty_generate(json)
      end # end to_json method
      
      def define_role_argument (role_file)
         role_f = role_file.gsub(/\.rb$/, "")
         return {
         				'scriptkey' => 'ROLE_FILE',
                        'scriptvalue' => role_f,
                        'scriptdefaultvalue' => role_f
                 }
      end # end define_role_argument method
      
      def add_args(rolefile, model)
         args = []

         args << define_role_argument(rolefile)
         model_role = "ConfigRole"
         model.flatten_children.select { |a| a.kind_of?(Automation) && a.type == :chef_automation }.each do |automation|
            automation.amd.each do |id, attribute|
               # puts "#{attribute.class}:::#{id} --> #{attribute}"
               if id == :role
                  model_role = attribute.value
               end
               if attribute.name.nil? 
                  @@logger.warn("Automation attribute '#{id} has no name; skipping")
               elsif attribute.mutable
                  if attribute.value.nil?
                  	attribute.value=""
                  elsif attribute.password?
                  	attribute.value=""
                  	# This is done because weaver passwords are encrypted, and
                  	# the simple chef packager doesn't currently have a way to
                  	# decrypt them later
                  elsif (attribute.value.kind_of?(Array))
                    # IWD doesn't handle values with [ ] right now, so
                    # change these to "" as a work-around
                    attribute.value=""
                  end
                  arg = {
                  		'scriptkey' => attribute.id,
                        'scriptvalue' => attribute.value,
                        'scriptdefaultvalue' => attribute.value
                          }

                  args << arg
               end # if ... elsif 'true' == attribute[:mutable]
            end # automation.amd.each
         end # model.flatten_children.select
         
         return args
      end # end add_args method
      
    end # end IwdScriptGenerator class
  end # end Iwd module
end # end Weaver module
