require 'erubis'
require 'string'
require 'json'
require 'devops_deployment_lib/providers/iwd_virtual_system_client'
require 'weaver/core/weaver_error'
require 'weaver/iwd/iwd_request_generator'

module Weaver
  module Iwd
    # Enables the user to create a set of json files that return the requests related to Virtual System Patterns.
    # Given a pattern id, it will generate json files for the pattern, each of the parts, the virtual image
    # for each part, each of the script packages in the parts, the environment profiles and the cloud groups.
    # The json files include all of the response bodies that are used
    # to generate an IWD infrastructure.
    # In addition to the json files,
    # a file is saved that has the stub requests that can be copied
    # for use into a ruby spec file or class
    #
    # ===IWD Pattern to Webmock json files
    #
    # Author:: Tova Roth
    #
    class IwdWeaverWebMockGenerator

      # IWD hostname or fully-qualified hostname
      attr_reader :host
      
      # IWD user name
      attr_reader :username
      
      # IWD password
      attr_reader :password
      
      # If true, a valid SSL certificate will be required
      attr_reader :verify_ssl_cert
      
      # Creates a new IwdWeaverWebMockGenerator for a given IWD server that enables
      # the user to create WebMock json files.
      #
      # * *Args*    :
      #   - +host+ -> the IWD server hostname (String)
      #   - +username+ -> the user for the server (String)
      #   - +password+ -> the password for the server (String)
      #   - +verify_ssl_cert+ -> flag to indicate whether to validate SSL
      #                          certificates of the server (Boolean; default: false)
      # * *Returns* :
      #   - A newly created IwdPatternAdapter
      #
      def initialize(host, username, password, verify_ssl_cert = false)
        @host = host
        @username = username
        @password = password
        @verify_ssl_cert = verify_ssl_cert
        @client = DevopsDeploymentLib::IwdVirtualSystemClient.new(host, username, password)
        if !verify_ssl_cert
          @client.ssl_verify_mode=(OpenSSL::SSL::VERIFY_NONE)
        end
        @files_written = []
      end

      def generate_webmocks(iwd_pattern_id, path)

        # write pattern json to a file
        json = @client.describe_pattern(iwd_pattern_id)
        filename = "pattern_#{iwd_pattern_id}_response.json"
        write_json_to_file(json, filename, path)
        
        # write version to a file
        version = @client.describe_version
        filename = "pattern_#{iwd_pattern_id}_version_response.json"
        write_json_to_file(version, filename, path)

        # write cloud groups to a file
        cloud_groups = @client.describe_cloud_groups
        filename = "cloud_groups_response.json"
        write_json_to_file(cloud_groups, filename, path)

        #      write environment profiles to a file
        environment_profiles = @client.describe_environment_profiles
        filename = "environment_profiles_response.json"
        write_json_to_file(environment_profiles, filename, path)
        
        # write ip groups to a file
        ip_groups = @client.describe_ip_groups
        filename = "ip_groups_response.json"
        write_json_to_file(ip_groups, filename, path)

        # Map each part in JSON to an iwd_node in Weaver
        json['parts'].each do |part_json|
          part_id =  part_json['id']

          # write part to a file
          part_json_details = @client.describe_pattern_part(iwd_pattern_id, part_id)
          filename = "pattern_#{iwd_pattern_id}_part_#{part_id}_response.json"
          write_json_to_file(part_json_details, filename, path)
          
          # write part scripts to a file
          part_scripts = @client.describe_part_scripts(iwd_pattern_id, part_id)
          filename = "pattern_#{iwd_pattern_id}_part_#{part_id}_scripts_response.json"
          write_json_to_file(part_scripts, filename, path)

          # write script packages to files
          write_script_packages(iwd_pattern_id, part_id, path)

          # write virtual image to a file
          virt_img_location = part_json['virtualimage']  # get image URL and get id: "virtualimage"=>"/resources/virtualImages/28"
          if !virt_img_location.nil?
            virt_image_id = virt_img_location.slice(virt_img_location.rindex('/') + 1,  virt_img_location.length)   # remove the /resouces/.../ part
            vi_json = @client.describe_virtual_image(virt_image_id)
            filename = "virtual_image_#{virt_image_id}_response.json"
            write_json_to_file(vi_json, filename, path)
            
            # write virtual image topologies to a file
            virt_image_part_id = part_json_details['resid']
            cloud_groups = @client.describe_virtual_image_topologies(virt_image_id, virt_image_part_id)
            filename = "virtual_image_#{virt_image_id}_part_#{virt_image_part_id}_topologies_response.json"
            write_json_to_file(cloud_groups, filename, path)
          end
         

        end

      end

      private

      # Writes script packages to json file.
      # * *Args*    :
      #   - +pattern_id+ -> the id of the pattern
      #   - +part_id+ -> the id of the pattern part
      def write_script_packages(pattern_id, part_id, path)
        scripts = @client.describe_part_scripts(pattern_id, part_id)
        unless scripts.nil?
          scripts.each do |script_json|
            script_id = script_json['scriptId']
            script_def_json = @client.describe_script_package(script_id)
            if script_def_json
              filename = "script_package_#{script_id}_response.json"
              write_json_to_file(script_def_json, filename, path)
            end
          end
        end
      end

      # Writes json to a file. Creates a backup if the file already exists.
      # * *Args*    :
      #   - +json+ -> the json response to write
      #   - +filename+ -> the filename without the path
      #   - +path+ -> the path, or nil for current folder
      def write_json_to_file(json, filename, path)

        full_filename = filename
        if path
          full_filename = File.join(path, filename)
        end

        unless @files_written.include? full_filename
          puts "Writing #{filename}..."
          # create a backup copy if file already exists
          if File.exists?(full_filename)
            puts("Creating:", "File '#{filename}' already exists. Creating backup '#{filename}.saved'")
            FileUtils.copy(full_filename, "#{full_filename}.saved")
          end

          #       jsonString = json.to_json
          prettyJson = JSON.pretty_generate(json)
          File.open(full_filename, 'w') {|f| f.write(prettyJson) }
          @files_written.push(full_filename)
        end
      end

    end

  end
end