require 'weaver/core/logging'

module Weaver
  # \Weaver::Transforms provides operations to transform a Weaver model into another model.
  module Transforms
    include Weaver::Core

    # Implements duplicating a Weaver object deeply
    # (#clone via {Weaver::Core::Abstract#initialize_copy} *is* deep, but only for contained entities, 
    # not references such as {Weaver::Core::Node#realized_by})
    #
    # Author:: Ed Snible
    class Duplicate
      include Weaver::Core

      # Create a deep copy of +entity+.
      # * +entity+ A Weaver::Core::Base   
      def self.duplicate(entity)
        duplicate = Duplicate.new(entity)
        return duplicate.deep_entity
      end

      # @private
      # :stopdoc:
      def deep_entity
        return @entity
      end
      # :startdoc
      
      protected
      
      # @private
      # :stopdoc:
      def initialize(orig_entity)
        
        # Create a deep copy
        @entity = orig_entity.dup
        
        # If we are duplicate a top-level environment Topology, we should duplicate the Root as well because
        # otherwise there is a path back to the original
        if @entity.scope.kind_of?(Root)
          @entity.scope = orig_entity.scope.clone()
          @entity.scope.topologies[@entity.id] = @entity
          @entity.scope.components = {}
        end
        
        @old_to_new = {}
        build_mapping(@entity, orig_entity)
        
#        @@logger.debug "Duplicate.initialize built mapping ="
#        @old_to_new.each { |orig, replacement|
#          @@logger.debug "#{orig.fully_qualified_id} (object_id #{orig.object_id}) => (object_id #{replacement.object_id})"
#        }
        
        @entity.update_references(@old_to_new)
      end
      
      # RECURSIVE
      def build_mapping(new_entity, orig_entity)
        @old_to_new[orig_entity] = new_entity
        orig_children = orig_entity.contained_entities
        new_children = new_entity.contained_entities 
        new_children.each do |id, new_child|
          raise WeaverError.new("Missing child #{id.inspect} on #{orig_entity.id.inspect}") if orig_children[id].nil? 
          build_mapping(new_child, orig_children[id]) 
        end
      end
      # :startdoc
      
    end      
  end
end
