################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2013. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require 'aquarium'
require 'weaver_coordination'
require 'open3'
require 'json'
require 'weaver/continuous_validation'
require 'weaver/continuous_validation/service_client'

module Weaver
  module Analytics
        
    # Performs AOP-based instrumentation of the Chef runtime
    # for determining how individual recipes change the runtime state
    # in order to support learning of continuous validation rules and
    # recipe fingerprinting (xprinting).
    #
    # Author:: Fabio Oliveira
    #
    class ChefRecipeInterceptor
      include Weaver::Coordination::Logging
      include ::Aquarium::Aspects
       
      IGNORED_RECIPES = ["weaver::coordination", "weaver::analytics", "chef_handler::default", "weaver::origami_crawler"]
      XPRINT_CMD = 'xprint.py'
      SNAP_CMD = 'weaver_snapshot.py'
      
      def initialize(node_id, fq_node_id, service_hostname, service_port, weaver_project, deployment_id, xprint_install_dir)
        @node_id = node_id
        @fq_node_id = fq_node_id
        @service_hostname = service_hostname
        @service_port = service_port
        @weaver_project = weaver_project
        @deployment_id = deployment_id
        @xprint_install_dir = xprint_install_dir
        @current_recipe = nil
        @snap_list = []
          
        @service_client = Weaver::ContinuousValidation::ServiceClient.new(@service_hostname, @service_port) 
      end
      
      def register_resource_hook()
        Aspect.new(:around, :calls_to => :run_action, :for_types => [ Chef::Runner ]) do |jp, obj, *args|
          cookbook_name = args[0].cookbook_name
          resource_type = args[0].class
          resource_name = args[0].resource_name
          recipe_name = args[0].recipe_name
          @@logger.debug "Weaver intercepted jp = #{jp.method_name}"
          @@logger.debug "  Resource '#{resource_name}' of type #{resource_type}"
          @@logger.debug "    belonging to recipe '#{recipe_name}' of cookbook '#{cookbook_name}'"
          next_recipe = "#{cookbook_name}::#{recipe_name}" 
          
          if @current_recipe != next_recipe 
            # A new recipe will be executed by Chef
            @@logger.debug "NEXT RECIPE: Weaver detected that the recipe '#{next_recipe}' will start next."
            capture_state(@current_recipe, next_recipe)
            @current_recipe = next_recipe 
          end
          
          jp.proceed
        end
      end
      
      def register_converge_hook()
        Aspect.new(:after, :calls_to => :converge, :for_types => [ Chef::Runner ]) do |jp, obj, *args|
          # The last resource of the last recipe has been processed. Convergence is finished.
          @@logger.debug "Weaver detected that the execution of the last recipe '#{@current_recipe}' has ended. The convergence phase is finished."
          capture_state(@current_recipe, nil)
          
          @@logger.debug "Computing xprint between first and last snapshots. First = '#{@snap_list.first}'; Last = '#{@snap_list.last}'"
          wmd_file_name = "/tmp/overall.json"
          capture_weaver_metadata("overall", wmd_file_name)
          xprint(@snap_list.first, @snap_list.last, wmd_file_name)
          generate_continuous_validation_rules('/tmp/xp.json')
        end
      end
      
      protected
      
      def snaphsot(snapshot_name)
        cmd = 'python'
        snap_cmd = "#{@xprint_install_dir}/#{SNAP_CMD}"
        snapId = snapshot_name
        
        @@logger.debug "calling command #{cmd} #{snap_cmd} #{snapId} ..."
        o,e,s = Open3.capture3(cmd, snap_cmd, snapId)
        if s.success?
          @@logger.debug "successfully completed with ouput: "
          @@logger.debug "#{o}"
          @@logger.debug "#{e}"
          return true
        else
          @@logger.debug "snapshot terminated with error: "
          @@logger.debug "#{e}"
          @@logger.debug "#{o}"
          return false
        end 
      end # snapshot()
      
      def xprint(s1_id, s2_id, wmd_file)
        if s1_id.nil? or s2_id.nil? 
          @@logger.error "xprint called with nil snapshot id"
          return false
        end
        cmd = 'python'
        xprint_cmd = "#{@xprint_install_dir}/#{XPRINT_CMD}"
        @@logger.debug "calling command #{cmd} #{xprint_cmd} #{s1_id} #{s2_id} #{wmd_file} ..."
        o,e,s = Open3.capture3(cmd, xprint_cmd, s1_id, s2_id, wmd_file)
        if s.success?
          @@logger.debug "successfully completed xprint with output:"
          @@logger.debug "#{o}"
          @@logger.debug "#{e}"
          return true
        else
          @@logger.error "xprint failed with error:"
          @@logger.error "#{e}"
          @@logger.error "#{o}"
          return false
        end
      end # xprint()
      
      def capture_weaver_metadata(recipe_name, file_name)
        wmodel = {"weaver_project" => @weaver_project, \
                  "fq_node_id" => @fq_node_id, \
                  "recipe_name"=> recipe_name \
                }
        # FIXME: how to extract and pass the args of are recipe?
        params = []
        wmd = {"weaver" => {"weavermodel" => wmodel, "params" => params} }
        
        File.open(file_name,"w") do |f|
          f.write(wmd.to_json)
        end  
      end # end capture_weaver_metadata()
      
      
      # Example: Ignored recipes: "I*" 
      # R1 -> I1 -> R2
      # we would like to do:
      # snapshot() -> R1 -> snapshot(), xprint(), -> I1 -> snapshot() -> R2 -> snapshot(), xprint()
      ###
      def capture_state(previous_recipe, next_recipe)  
        if previous_recipe.nil?
          prn = 'FIRST_RECIPE'
        else
          prn = previous_recipe  
        end
        if next_recipe.nil?
          nrn = 'LAST_RECIPE'
        else
          nrn = next_recipe
        end
        # if previous_recipe.nil?, this is the first recipe.
        # if next_recipe.nil?, we are done with the last recipe.
        capture_state_flag =    (!IGNORED_RECIPES.include?(previous_recipe) && !previous_recipe.nil?) 
        capture_state_flag ||=  (!IGNORED_RECIPES.include?(next_recipe) && !next_recipe.nil?)
        if capture_state_flag
          @@logger.debug "Weaver will capture the runtime state now: previous_recipe = '#{prn}'; next_recipe = '#{nrn}'"
          snap_name = "#{prn}_x2x_#{nrn}"
          snap_success = snaphsot(snap_name)
          if snap_success
            @@logger.debug "Weaver completed snapshot : #{snap_name}"
            @snap_list << snap_name
          end
          # xprint
          if !IGNORED_RECIPES.include?(previous_recipe) && snap_success
            @@logger.debug "Weaver capturing metadata for recipe : #{prn} ..."
            wmd_file_name = "/tmp/#{prn}.json"
            capture_weaver_metadata(prn, wmd_file_name)
            @@logger.debug "Weaver deriving xprint(#{@snap_list[-1]}, #{@snap_list[-2]}) ..."
            xprint(@snap_list[-1], @snap_list[-2], wmd_file_name)
          end # end if not ignored recipe
        end # end if capture_state_flag
      end # end capture_state()
      
      
      # Generates one or more continuous validation rules. 
      # +xprint_filename+: Name of a file containing an xprint corresponding to the diff between the first and last snapshots
      #
      def generate_continuous_validation_rules(xprint_filename)
        xp_hash = JSON.parse(File.read(xprint_filename))
        generate_check_process_rules(xp_hash)
        generate_check_port_rule(xp_hash)
      end
      
      def generate_check_process_rules(xprint_hash)
        check_process_rules = Hash.new {|h,k| h[k] = []}

        xp_regex = Regexp.new(Regexp.escape("#{@xprint_install_dir}/#{XPRINT_CMD}"))
        snap_regex = Regexp.new(Regexp.escape("#{@xprint_install_dir}/#{SNAP_CMD}"))

        # Iterate over processes
        xprint_hash['process'].each do |pid, metadata|
          if metadata['dtype'] == 'added'
            # This process was created
            skip_process = false
            metadata['command'].each do |line|      
              if line =~ xp_regex || line =~ snap_regex 
                # Skip this process, as it is one of ours
                @@logger.debug "Skipping process because of command line substring: '#{line}'"
                skip_process = true
                break
              end
            end
            next if skip_process

            user = metadata['user']
            full_command_line = metadata['command'].join(' ')
            if check_process_rules[full_command_line].include?(user)
              # We have already generated a rule to check for this process being run as this user
              # Skip it
              next
            end        

            # It seems that we need to generate a continuous validation rule to check for this process
            check_process_rules[full_command_line] << user
            @@logger.debug "Requesting generation of a continuous validation rule for process: user = #{user}; full_command = #{full_command_line}"
            @service_client.generate_check_process_rule(@weaver_project, @node_id, pid, user, full_command_line)
          end
        end
      end
      
      def generate_check_port_rule(xprint_hash)
        ports = []
        # Iterate over network
        xprint_hash['network'].each do |port, metadata|
          if metadata['dtype'] == 'added'
            # This port was opened
            ports << port
          end
        end

        unless ports.size == 0
          @@logger.debug "Requesting generation of a continuous validation rule for ports: #{ports.inspect}"
          @service_client.generate_check_port_rule(@weaver_project, @node_id, ports)
        end
      end
    end # end class ChefRecipeInterceptor
    
  end # end module Bootstrap
end # end module Weaver